/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.app.Activity
import android.app.Application
import android.content.Context
import android.net.Uri
import android.util.Log
import androidx.annotation.NonNull
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import com.google.gson.JsonObject
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.scanbot.sap.IScanbotSDKLicenseErrorHandler
import io.scanbot.sap.SdkFeature
import io.scanbot.sap.Status
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.ScanbotSDKInitializer
import io.scanbot.sdk.barcode.ScanbotBarcodeDetector
import io.scanbot.sdk.barcode.entity.BarcodeDocumentFormat
import io.scanbot.sdk.barcode.entity.BarcodeFormat
import io.scanbot.sdk.core.contourdetector.ContourDetector
import io.scanbot.sdk.flutter.camera.CameraFactory
import io.scanbot.sdk.flutter.utils.StorageUtils
import io.scanbot.sdk.flutter.utils.getBitmapFromContentUri
import io.scanbot.sdk.flutter.utils.getBitmapFromFileUri
import io.scanbot.sdk.persistence.CameraImageFormat
import io.scanbot.sdk.persistence.PageFileStorage
import io.scanbot.sdk.persistence.PageStorageSettings
import io.scanbot.sdk.persistence.fileio.AESEncryptedFileIOProcessor
import kotlinx.coroutines.*
import java.io.File
import java.text.SimpleDateFormat
import java.util.*
import kotlin.coroutines.CoroutineContext


class ScanbotSdkPlugin : FlutterPlugin,
    ActivityAware,
    MethodCallHandler,
    CoroutineResultHandler(),
    CoroutineScope,
    ISdkSubpluginManager by SdkSubpluginManager() {

    private var job: Job = Job()
    private lateinit var scanbotSDK: ScanbotSDK
    private var channel: MethodChannel? = null
    var batchBarcodeCallbackChannel: MethodChannel? = null
    private val gson = GsonBuilder()
        .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
        .registerTypeAdapter(kotlin.ByteArray::class.java, ByteArrayToBase64TypeAdapter())
        .create()

    private var currentActivity: Activity? = null
    private var appContext: Context? = null
    override val coroutineContext: CoroutineContext
        get() = job + Dispatchers.Default

    override fun onMethodCall(call: MethodCall, result: Result) {
        launch(Dispatchers.IO) {
            when (call.method) {
                "initScanbotSdk" -> {
                    appContext?.let { initScanbotSdk(it, call, result) }
                }
                "getLicenseStatus" -> {
                    isLicenseValid(call, result)
                }
                "detectBarcodesOnImages" -> {
                    currentActivity?.let { detectBarcodesOnImages(it, call, result) }
                }
                "estimateBlurOnPage" -> {
                    currentActivity?.let { estimateBlurOnPage(it, call, result) }
                }
                else -> {
                    if (!callMethod(call, result)) {
                        withContext(Dispatchers.Main) { result.notImplemented() }
                    }
                }
            }
        }
    }

    private suspend fun initScanbotSdk(
        context: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                val licenseKey = call.argument<String>("licenseKey") ?: ""
                val scanFormat = call.argument<String>("imageFormat") ?: ""
                val storageBaseDirectory = call.argument<String>("storageBaseDirectory")
                val imageQuality = call.argument<Int>("imageQuality") ?: 80
                val previewTargetMax = call.argument<Int>("previewTargetMax") ?: 900
                val loggingEnabled = call.argument<Boolean>("loggingEnabled") ?: false
                val encryptionParams = call.argument<Json>("encryptionParameters")

                val documentDetectorMode = call.argument<String>("documentDetectorMode")?.let {
                    ContourDetector.Type.valueOf(it)
                }

                val useCameraX = call.argument<Boolean>("useCameraX") ?: false
                val allowGpuAcceleration = call.argument<Boolean>("allowGpuAcceleration") ?: true
                val allowXnnpackAcceleration =
                    call.argument<Boolean>("allowXnnpackAcceleration") ?: true

                val imageFormat = if (scanFormat.contains("PNG")) {
                    CameraImageFormat.PNG
                } else {
                    CameraImageFormat.JPG
                }

                val scanbotSDKInitializer = ScanbotSDKInitializer()

                if (!storageBaseDirectory.isNullOrEmpty()) {
                    val uri = Uri.parse(storageBaseDirectory)
                    val customStorageBaseDirectory = File(uri.path)
                    customStorageBaseDirectory.mkdirs()
                    if (customStorageBaseDirectory.isDirectory.not()) {
                        methodResult.error(
                            "IOException",
                            "Specified storageBaseDirectory is not valid: $uri",
                            null
                        )
                    }
                    if (!customStorageBaseDirectory.exists()) {
                        customStorageBaseDirectory.mkdirs()
                    }
                    scanbotSDKInitializer.sdkFilesDirectory(
                        context as Application,
                        customStorageBaseDirectory
                    )
                    StorageUtils.setCustomStorageBaseDirectory(customStorageBaseDirectory)
                }
                documentDetectorMode?.let {
                    scanbotSDKInitializer.contourDetectorType(it)
                }
                var encryptionEnabled = false
                encryptionParams?.let {
                    val password = it["password"] as? String
                    val encryptorMode = (it["mode"] as? String)?.let {
                        AESEncryptedFileIOProcessor.AESEncrypterMode.valueOf(it)
                    } ?: AESEncryptedFileIOProcessor.AESEncrypterMode.AES256
                    if (password != null) {
                        val encryptor = AESEncryptedFileIOProcessor(password, encryptorMode)
                        scanbotSDKInitializer.useFileEncryption(true, encryptor)
                        encryptionEnabled = true
                    } else {
                        methodResult.error(
                            "RuntimeException",
                            "Encryption password not set",
                            null
                        )
                    }
                }
                scanbotSDKInitializer
                    .licenceErrorHandler(IScanbotSDKLicenseErrorHandler { status, feature, message ->
                        val error = gson.toJson(
                            FlutterPluginLicenseError(
                                status, feature, message
                                    ?: ""
                            )
                        )
                        channel?.invokeMethod("onLicenseError", error)
                    })
                    .withLogging(loggingEnabled)
                    .usePageStorageSettings(
                        PageStorageSettings.Builder()
                            .imageFormat(imageFormat)
                            .imageQuality(imageQuality)
                            .previewTargetMax(previewTargetMax)
                            .build()
                    )
                    .prepareOCRLanguagesBlobs(true)
                    .prepareMRZBlobs(true)
                    .preparePayFormBlobs(true)
                    .license(context as Application, licenseKey)
                    .useCameraXRtuUi(useCameraX)
                    .allowGpuAcceleration(allowGpuAcceleration)
                    .allowXnnpackAcceleration(allowXnnpackAcceleration)
                    .initialize(context)

                scanbotSDK = ScanbotSDK(context)
                clearPlugins()

                currentActivity?.let {
                    attachActivityAwareSubplugins(it)
                }

                attachContextAwareSubplugins(context, encryptionEnabled)

                if ("" != licenseKey.trim() && "null" != licenseKey.toLowerCase()) {
                    methodResult.success("Scanbot SDK initialized.")
                } else {
                    methodResult.success(
                        "Scanbot SDK initialized. Trial mode activated. You can now test all features for 60 seconds."
                    )
                }
            } catch (e: Throwable) {
                Log.e("SDK", e.toString())
                methodResult.error(e::class.java.simpleName, e.message, null)
            }
        }
    }

    private suspend fun estimateBlurOnPage(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val blurEstimator = scanbotSDK.createBlurEstimator()
                val pageJson = call.argument<Json>("page") ?: Json()
                val page = pageJson.toPage(gson)
                val pageFileStorage = scanbotSDK.createPageFileStorage()
                val documentUri =
                    pageFileStorage.getImageURI(page.pageId, PageFileStorage.PageFileType.DOCUMENT)
                val originalUri =
                    pageFileStorage.getImageURI(page.pageId, PageFileStorage.PageFileType.ORIGINAL)
                val document = File(documentUri.path ?: "")
                val original = File(originalUri.path ?: "")
                val file = when {
                    document.exists() -> document
                    original.exists() -> original
                    else -> null
                }

                file?.let {
                    val bitmap = scanbotSDK.fileIOProcessor().readImage(it)

                    bitmap?.let {
                        val blurValue = blurEstimator.estimateInBitmap(it, 0)
                        blurValue?.let {
                            withContext(Dispatchers.Main) { methodResult.success(it) }
                        }
                    }
                        ?: withContext(Dispatchers.Main) {
                            methodResult.error(
                                "IllegalArgumentException",
                                "bitmap not found",
                                null
                            )
                        }

                }
                    ?: methodResult.error(
                        "IllegalArgumentException",
                        "Page does not have any available images to estimate blurriness",
                        null
                    )


            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "IllegalStateException",
                        e.message,
                        null
                    )
                }
            }
        }
    }

    private fun attachActivityAwareSubplugins(activity: Activity) {
        attachSubplugin(
            ScanbotSdkUiPlugin(
                activity,
                batchBarcodeCallbackChannel,
                this@ScanbotSdkPlugin,
                scanbotSDK
            )
        )
    }

    private fun attachContextAwareSubplugins(context: Context, encryptionEnabled: Boolean) {
        attachSubplugin(
            PageFileStoragePlugin(
                context,
                this@ScanbotSdkPlugin,
                scanbotSDK
            )
        )

        attachSubplugin(
            PageProcessorPlugin(
                context,
                this@ScanbotSdkPlugin,
                scanbotSDK,
                encryptionEnabled
            )
        )
    }

    private fun detachContextAwareSubplugins() {
        detachSubpluginByType(PageFileStoragePlugin::class.java)
        detachSubpluginByType(PageProcessorPlugin::class.java)
    }

    private fun detachActivityAwareSubplugins() {
        detachSubpluginByType(ScanbotSdkUiPlugin::class.java)
    }

    private suspend fun isLicenseValid(
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                if (::scanbotSDK.isInitialized) {
                    val json = JsonObject()
                    json.addProperty("isLicenseValid", scanbotSDK.licenseInfo.isValid)
                    json.addProperty("status", scanbotSDK.licenseInfo.status.name)
                    val time = scanbotSDK.licenseInfo.expirationDate?.time
                    time?.let { date ->
                        val sdf: SimpleDateFormat = SimpleDateFormat(
                            "yyyy-MM-dd'T'HH:mm:ss.SSS",
                            Locale.ENGLISH
                        )
                        sdf.timeZone = TimeZone.getTimeZone("UTC")
                        json.addProperty("expirationDate", sdf.format(date))
                    }
                    methodResult.success(json.toString())
                } else {
                    methodResult.error(
                        "SDKNotInitialisedException",
                        "need to initialise sdk with licence",
                        null
                    )
                }
            } catch (e: Throwable) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                    "Sdk exception",
                    e.localizedMessage,
                    null
                )
            }
        }
    }

    override fun onDetachedFromActivity() {
        currentActivity = null
        detachActivityAwareSubplugins()
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        val activity = binding.activity
        binding.addActivityResultListener(this)
        currentActivity = activity
        if (::scanbotSDK.isInitialized) {
            attachActivityAwareSubplugins(activity)
        }
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        val activity = binding.activity
        binding.addActivityResultListener(this)
        currentActivity = activity
        if (::scanbotSDK.isInitialized) {
            attachActivityAwareSubplugins(activity)
        }
    }

    override fun onDetachedFromActivityForConfigChanges() {
        currentActivity = null
        detachActivityAwareSubplugins()
    }

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        appContext = flutterPluginBinding.applicationContext
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "scanbot_sdk")
        batchBarcodeCallbackChannel =
            MethodChannel(flutterPluginBinding.binaryMessenger, "batch_barcode_data_mapper")
        channel?.setMethodCallHandler(this);

        flutterPluginBinding
                .platformViewRegistry
                .registerViewFactory(
                        "scanbot_sdk_camera", CameraFactory(flutterPluginBinding.binaryMessenger))

    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        job.cancel()
        clearPlugins()
    }

    /// common method used for detecting the barcodes
    private suspend fun detectBarcodesOnImages(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        //too add high sensitivity mode to and remove rotation
        withContext(Dispatchers.Default) {
            try {
                val barcodeDetector = setUpBarcodeDetector(call);
                val uris = call.argument<ArrayList<String>>("uris")
                if (uris == null || uris.count() == 0) {
                    methodResult.error("IllegalArgumentException", "Need to setup file uri =", null)
                } else {
                    var barcodeResults = ArrayList<Map<String, Any>>();
                    for (uriStringItem in uris) {
                        val uri = Uri.parse(uriStringItem)
                        val file = File(uri.path)
                        val bitmap = when (uri.scheme) {
                            "content" -> {
                                getBitmapFromContentUri(activity.contentResolver, uri)
                            }
                            "file" -> {
                                getBitmapFromFileUri(file)
                            }
                            else -> {
                                if (file.exists()) {
                                    getBitmapFromFileUri(file)
                                } else {
                                    barcodeResults.add(
                                        getBarcodeDetectionError(
                                            uriStringItem,
                                            "Unknown file path format."
                                        )
                                    );
                                    return@withContext;
                                }
                            }
                        }

                        bitmap?.let {
                            val barcodeData = barcodeDetector.detectFromBitmap(bitmap, 0)
                            barcodeData?.let { barcodeDataResult ->
                                var successData = mapOf<String, Any>(
                                    "barcodeImageURI" to uriStringItem,
                                    "barcodeItems" to barcodeData.barcodeItems
                                );

                                barcodeResults.add(successData);
                            }
                            if (barcodeData == null || barcodeData.barcodeItems?.count() == 0) {
                                barcodeResults.add(
                                    getBarcodeDetectionError(
                                        uriStringItem,
                                        "No barcodes found."
                                    )
                                );
                            }
                        }

                        if (bitmap == null) {
                            barcodeResults.add(
                                getBarcodeDetectionError(
                                    uriStringItem,
                                    "bitmap not found."
                                )
                            );
                        }
                    }

                    var result = mapOf("barcodeResults" to barcodeResults);
                    methodResult.success(gson.toJson(result));
                }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error("IllegalStateException", e.message, null)
                }
            }
        }
    }

    /// get the error message payload
    private fun getBarcodeDetectionError(uri: String, errorMessage: String): Map<String, Any> {
        return mapOf<String, Any>(
            "barcodeImageURI" to uri,
            "error" to errorMessage,
            "barcodeItems" to ArrayList<String>()
        );
    }

    /// set up the barcode detector object and returns the same.
    private fun setUpBarcodeDetector(call: MethodCall): ScanbotBarcodeDetector {
        var barcodeDetector = scanbotSDK.createBarcodeDetector()
        when {
            call.hasArgument("barcodeFormats") -> {
                val arguments = call.arguments as Json
                val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
                val acceptedBarcodeFormats =
                    barcodeFormats.map { BarcodeFormat.valueOf(it) }
                barcodeDetector.modifyConfig {
                    this.setBarcodeFormats(acceptedBarcodeFormats)
                }
            }
            call.hasArgument("acceptedDocumentFormats") -> {
                val arguments = call.arguments as Json
                val acceptedDocumentFormats =
                    arguments["acceptedDocumentFormats"] as ArrayList<String>
                val documentFormats =
                    acceptedDocumentFormats.map { BarcodeDocumentFormat.valueOf(it) }
                barcodeDetector.modifyConfig {
                    this.setAcceptedDocumentFormats(documentFormats)
                }
            }
            else -> {
                barcodeDetector.modifyConfig {
                    this.setBarcodeFormats(BarcodeFormat.values().toList())
                }
            }
        }
        return barcodeDetector;
    }
}

class FlutterPluginLicenseError(val status: Status, val feature: SdkFeature, val message: String)
