/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.app.Activity
import android.net.Uri
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.Result
import io.scanbot.genericdocument.entity.GenericDocument
import io.scanbot.hicscanner.model.HealthInsuranceCardRecognitionResult
import io.scanbot.mrzscanner.model.MRZRecognitionResult
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.barcode.entity.BarcodeFormattedData
import io.scanbot.sdk.barcode.entity.BarcodeItem
import io.scanbot.sdk.barcode.entity.BarcodeScanningResult
import io.scanbot.sdk.barcode.entity.FormattedBarcodeDataMapper
import io.scanbot.sdk.persistence.Page
import io.scanbot.sdk.ui.result.ResultWrapper
import io.scanbot.sdk.ui.view.barcode.BarcodeScannerActivity
import io.scanbot.sdk.ui.view.barcode.InternalBarcodeScannerActivity
import io.scanbot.sdk.ui.view.barcode.batch.BatchBarcodeScannerActivity
import io.scanbot.sdk.ui.view.base.RtuConstants
import io.scanbot.sdk.ui.view.camera.DocumentScannerActivity
import io.scanbot.sdk.ui.view.edit.CroppingActivity
import io.scanbot.sdk.ui.view.genericdocument.GenericDocumentRecognizerActivity
import io.scanbot.sdk.ui.view.hic.HealthInsuranceCardScannerActivity
import io.scanbot.sdk.ui.view.licenseplate.LicensePlateScannerActivity
import io.scanbot.sdk.ui.view.licenseplate.entity.LicensePlateScannerResult
import io.scanbot.sdk.ui.view.mrz.MRZScannerActivity
import io.scanbot.sdk.ui.view.nfc.NfcPassportScannerActivity
import io.scanbot.sdk.ui.view.nfc.entity.NfcPassportScanningResult
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.withContext
import java.util.ArrayList
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException
import kotlin.coroutines.suspendCoroutine

class ScanbotSdkUiPlugin(
    val activity: Activity,
    var barcodeCallbackChannel: MethodChannel?,
    coroutineResultHandler: CoroutineResultHandler,
    val scanbotSDK: ScanbotSDK
) : ISdkSubplugin,
    ICoroutineActivityResultHandler by coroutineResultHandler {

    init {
        batchBarcodeCallbackChannel = barcodeCallbackChannel
    }

    companion object {
        @JvmStatic
        var batchBarcodeCallbackChannel: MethodChannel? = null

        @JvmStatic
        private val gson = GsonBuilder()
            .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
            .registerTypeAdapter(kotlin.ByteArray::class.java, ByteArrayToBase64TypeAdapter())
            .create()


        @JvmStatic
        private suspend fun getFormattedData(barcodeItem: BarcodeItem) =
            suspendCoroutine<BarcodeFormattedData> {
                val json = gson.toJson(barcodeItem)
                batchBarcodeCallbackChannel?.invokeMethod(
                    "getFormattedData",
                    json.toString(),
                    object : Result {
                        override fun success(result: Any?) {
                            val json = (result as Json)
                            val data = BarcodeFormattedData(
                                (json["title"] as? String)
                                    ?: barcodeItem.barcodeFormat.name,
                                (json["subtitle"] as? String) ?: barcodeItem.text ?: ""
                            );
                            it.resume(data)
                        }

                        override fun error(
                            errorCode: String,
                            errorMessage: String?,
                            errorDetails: Any?
                        ) {
                            it.resumeWithException(IllegalStateException(errorMessage))
                        }

                        override fun notImplemented() {
                            it.resumeWithException(IllegalStateException("Not yet implemented"))
                        }
                    })
            }
    }

    override suspend fun onMethodCall(call: MethodCall, result: Result): Boolean {
        when (call.method) {
            "startBarcodeScanner" -> {
                proceedWithBarcodeScanner(activity, call, result)
            }
            "startBatchBarcodeScanner" -> {
                proceedWithBatchBarcodeScanner(activity, call, result)
            }
            "startMrzScanner" -> {
                proceedMRZScanner(activity, call, result)
            }
            "startDocumentScanner" -> {
                proceedDefaultScanner(activity, call, result)
            }
            "startGenericDocumentRecognizer" -> {
                proceedGenericDocumentRecognizer(activity, call, result)
            }
            "startCroppingScreen" -> {
                proceedCroppingScreen(activity, call, result)
            }
            "startNfcPassportReader" -> {
                startNfcPassportReader(activity, call, result)
            }
            "startEhicScanner" -> {
                proceedEhicScreen(activity, call, result)
            }
            "startLicensePlateScanner" -> {
                proceedLicensePlateScreen(activity, call, result)
            }
            else -> {
                return false
            }
        }
        return true
    }

    private suspend fun proceedWithBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBarcodeScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) {
                methodResult.error(
                    e::class.java.simpleName,
                    e.message,
                    null
                )
            }
            return
        }
        val intent =
            BarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(barcodeData).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }

    private suspend fun proceedWithBatchBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBatchBarcodeScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error("ParseConfigException", "", null) }
            return
        }

        val arguments = call.arguments as Json
        val isNeedToSetCallback = (arguments["shouldUseBarcodeFormatter"] as? Boolean) ?: false
        val intent = if (isNeedToSetCallback) {
            BatchBarcodeScannerActivity.newIntent(
                activity,
                barcodeCameraConfiguration,
                FlutterInnerBarcodeFormatter::class.java
            )
        } else {
            BatchBarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        }


        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(it).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }

    class FlutterInnerBarcodeFormatter : FormattedBarcodeDataMapper {
        override fun decodeFormattedData(barcodeItem: BarcodeItem): BarcodeFormattedData {
            try {
                return runBlocking {
                    withContext(Dispatchers.Main) { getFormattedData(barcodeItem) }
                }
            } catch (e: Throwable) {
                return BarcodeFormattedData(barcodeItem.barcodeFormat.name, barcodeItem.text)
            }

        }
    }

    private suspend fun startNfcPassportReader(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {

            val configuration = try {
                readNFCPassportScannerConfiguration(call, activity)
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "ParseConfigException",
                        "",
                        null
                    )
                }
                return@withContext
            }
            val intent =
                NfcPassportScannerActivity.newIntent(activity, configuration)
            val activityResult = withContext(Dispatchers.Main) {
                launchIntent(activity, intent).await()
            }

            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {

                    val result =
                        it.data?.getParcelableExtra<NfcPassportScanningResult>(
                            RtuConstants.EXTRA_KEY_RTU_RESULT
                        )

                    result?.let {
                        val resultJson = gson.toJsonTree(result) as JsonObject
                        fixNfcResultForFlutter(resultJson, result.photo)
                        withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                    } ?: returnEmptyResult(methodResult)
                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }

        }
    }


    private suspend fun proceedEhicScreen(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {

            val configuration = try {
                readEhicScannerConfiguration(call, activity)
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "ParseConfigException",
                        "",
                        null
                    )
                }
                return@withContext
            }
            val intent =
                HealthInsuranceCardScannerActivity.newIntent(activity, configuration)
            val activityResult = withContext(Dispatchers.Main) {
                launchIntent(activity, intent).await()
            }

            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {

                    val result =
                        it.data?.getParcelableExtra<HealthInsuranceCardRecognitionResult>(
                            RtuConstants.EXTRA_KEY_RTU_RESULT
                        )

                    result?.let {
                        val resultJson = gson.toJsonTree(result) as JsonObject
                        withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                    } ?: returnEmptyResult(methodResult)
                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }

        }
    }

    private suspend fun proceedLicensePlateScreen(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {

            val configuration = try {
                readLicensePlateScannerConfiguration(call, activity)
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "ParseConfigException",
                        "",
                        null
                    )
                }
                return@withContext
            }
            val intent =
                LicensePlateScannerActivity.newIntent(activity, configuration)
            val activityResult = withContext(Dispatchers.Main) {
                launchIntent(activity, intent).await()
            }

            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {

                    val result =
                        it.data?.getParcelableExtra<LicensePlateScannerResult>(
                            RtuConstants.EXTRA_KEY_RTU_RESULT
                        )

                    result?.let {
                        val resultJson = gson.toJsonTree(result) as JsonObject
                        withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                    } ?: returnEmptyResult(methodResult)
                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }

        }
    }

    private suspend fun proceedCroppingScreen(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            val croppingConfiguration = try {
                readCroppingConfiguration(call, activity, gson)
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "ParseConfigException",
                        "",
                        null
                    )
                }
                return@withContext
            }
            val intent =
                CroppingActivity.newIntent(activity, croppingConfiguration)
            val activityResult = withContext(Dispatchers.Main) {
                launchIntent(activity, intent).await()
            }

            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {
                    val resultJson = JsonObject()

                    val page =
                        it.data?.getParcelableExtra<Page>(RtuConstants.EXTRA_KEY_RTU_RESULT)

                    page?.let {
                        val pageJson = page.toJson(gson)
                        insertPageUris(scanbotSDK, this, page.pageId, pageJson)
                        resultJson.add("page", pageJson)
                        withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                    } ?: returnEmptyResult(methodResult)
                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }

        }
    }

    private suspend fun returnEmptyResult(methodResult: Result) {
        withContext(Dispatchers.Main) {
            methodResult.error(
                "ERROR",
                "Empty camera result",
                null
            )
        }
    }

    private suspend fun proceedMRZScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val mrzScannerConfiguration = try {
            readMRZScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error("ParseConfigException", "", null) }
            return
        }
        val intent =
            MRZScannerActivity.newIntent(activity, mrzScannerConfiguration)
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }
        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<MRZRecognitionResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                barcodeData?.let {
                    val result = gson.toJson(it)
                    withContext(Dispatchers.Main) { methodResult.success(result) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }


    private suspend fun proceedDefaultScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val cameraConfiguration = try {
            readDocumentScannerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error("ParseConfigException", "", null) }
            return
        }
        val intent = DocumentScannerActivity.newIntent(
            activity,
            cameraConfiguration
        )
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }
        withContext(Dispatchers.Default) {
            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {

                    val resultJson = JsonObject()
                    val array = JsonArray()

                    it.data?.getParcelableArrayExtra(RtuConstants.EXTRA_KEY_RTU_RESULT)
                        ?.map { element ->
                            element as Page
                        }?.map { page ->
                            val pageJson = page.toJson(gson)
                            insertPageUris(scanbotSDK, this, page.pageId, pageJson)
                            pageJson
                        }?.forEach { json ->
                            array.add(json)
                        }?.let {
                            resultJson.add("pages", array)
                            withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                        } ?: returnEmptyResult(methodResult)

                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
        }
    }

    private suspend fun proceedGenericDocumentRecognizer(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val genericDocumentRecognizerConfiguration = try {
            readGenericDocumentRecognizerConfiguration(call, activity)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) { methodResult.error("ParseConfigException", "", null) }
            return
        }
        val intent = GenericDocumentRecognizerActivity.newIntent(
            activity,
            genericDocumentRecognizerConfiguration
        )
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }
        withContext(Dispatchers.Default) {
            activityResult?.let {
                if (it.resultCode == Activity.RESULT_OK) {
                    val resultWrappers: ArrayList<ResultWrapper<GenericDocument>>? =
                        it.data?.getParcelableArrayListExtra(RtuConstants.EXTRA_KEY_RTU_RESULT);

                    var resultJson = extractGenericDocumentRecognizerResultFromResultWrappers(
                        resultWrappers,
                        scanbotSDK,
                        activity.applicationContext
                    ).put("status", "OK")

                    withContext(Dispatchers.Main) { methodResult.success(resultJson.toString()) }
                } else {
                    dispatchCanceled(methodResult)
                }
            } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
        }
    }


    private suspend fun dispatchCanceled(methodResult: Result) {
        withContext(Dispatchers.Main) {
            methodResult.success("CANCELED")
        }
    }

}

