/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.content.Context
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Matrix
import android.media.ExifInterface
import android.net.Uri
import android.util.Base64
import com.google.gson.JsonObject
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.persistence.PageFileStorage
import io.scanbot.sdk.util.FileChooserUtils
import io.scanbot.sdk.util.bitmap.BitmapUtils
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.async
import java.io.IOException
import java.text.SimpleDateFormat
import java.util.*

suspend fun insertPageUris(
    scanbotSDK: ScanbotSDK,
    coroutineScope: CoroutineScope,
    pageId: String,
    pageJson: JsonObject
) {

    val originalImageFileUri = coroutineScope.async {
        scanbotSDK.createPageFileStorage()
            .getImageURI(pageId, PageFileStorage.PageFileType.ORIGINAL)
    }
    val documentImageFileUri = coroutineScope.async {
        scanbotSDK.createPageFileStorage()
            .getImageURI(pageId, PageFileStorage.PageFileType.DOCUMENT)
    }
    val originalPreviewImageFileUri = coroutineScope.async {
        scanbotSDK.createPageFileStorage()
            .getPreviewImageURI(pageId, PageFileStorage.PageFileType.ORIGINAL)
    }
    val documentPreviewImageFileUri = coroutineScope.async {
        scanbotSDK.createPageFileStorage()
            .getPreviewImageURI(
                pageId,
                PageFileStorage.PageFileType.DOCUMENT
            )
    }
    fillPageJsonWithUris(
        pageJson,
        originalImageFileUri.await(),
        documentImageFileUri.await(),
        originalPreviewImageFileUri.await(),
        documentPreviewImageFileUri.await()
    )
}

private suspend fun fillPageJsonWithUris(
    pageJson: JsonObject,
    originalImageFileUri: Uri,
    documentImageFileUri: Uri,
    originalPreviewImageFileUri: Uri,
    documentPreviewImageFileUri: Uri
) {
    pageJson.addProperty(
        "originalImageFileUri",
        (originalImageFileUri.toString())
    )
    pageJson.addProperty(
        "documentImageFileUri",
        (documentImageFileUri.toString())
    )
    pageJson.addProperty(
        "originalPreviewImageFileUri",
        (originalPreviewImageFileUri.toString())
    )
    pageJson.addProperty(
        "documentPreviewImageFileUri",
        (documentPreviewImageFileUri.toString())
    )
}


fun prepareImageForImport(imageUri: Uri, context: Context, fixOrientation: Boolean = true): Bitmap {

    when (imageUri.scheme) {
        "file" -> {
            val imagePath = FileChooserUtils.getPath(context, imageUri) ?: imageUri.path

            if (imagePath != null) {
                var bitmap: Bitmap? = BitmapUtils.decodeQuietly(imagePath, null)
                    ?: throw IOException("Could not load bitmap by given image file URI: $imageUri")
                if (fixOrientation) {
                    bitmap = fixOrientation(bitmap!!, imagePath)
                }
                return bitmap!!
            }
        }
        "content" -> {
            context.contentResolver.openInputStream(imageUri)
            return BitmapFactory.decodeStream(context.contentResolver.openInputStream(imageUri))
        }
        "data" -> { //special type created from flutter
            val bytes =
                Base64.decode(imageUri.toString().split("base64,").get(1), Base64.NO_WRAP)
            return BitmapFactory.decodeByteArray(bytes, 0, bytes.size)
        }
    }

    throw IOException("Cant read image: unknown type of the path");
}

private fun fixOrientation(bitmap: Bitmap, imagePath: String): Bitmap {

    val exif = ExifInterface(imagePath)
    val rotation = exif.getAttributeInt(
        ExifInterface.TAG_ORIENTATION,
        ExifInterface.ORIENTATION_NORMAL
    )
    val rotationInDegrees = exifToDegrees(rotation)
    val matrix = Matrix()
    if (rotation != 0) {
        matrix.preRotate(rotationInDegrees.toFloat())
        return Bitmap.createBitmap(bitmap!!, 0, 0, bitmap.width, bitmap.height, matrix, false)
    }


    return bitmap
}


private fun exifToDegrees(exifOrientation: Int): Int {
    return when (exifOrientation) {
        ExifInterface.ORIENTATION_ROTATE_90 -> 90
        ExifInterface.ORIENTATION_ROTATE_180 -> 180
        ExifInterface.ORIENTATION_ROTATE_270 -> 270
        else -> 0
    }
}

fun ScanbotSDK.getLicenseStatusJson(): JsonObject {
    val json = JsonObject()
    json.addProperty("isLicenseValid", licenseInfo.isValid)
    json.addProperty("status", licenseInfo.status.name)
    val time = licenseInfo.expirationDate?.time
    time?.let { date ->
        val sdf: SimpleDateFormat = SimpleDateFormat(
            "yyyy-MM-dd'T'HH:mm:ss.SSS",
            Locale.ENGLISH
        )
        sdf.timeZone = TimeZone.getTimeZone("UTC")
        json.addProperty("expirationDate", sdf.format(date))
    }
    val message = licenseInfo.licenseStatusMessage
    json.addProperty("message", message)
    return json
}
