/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.app.Activity
import android.content.Context
import android.graphics.Bitmap
import android.net.Uri
import android.util.Log
import com.google.gson.Gson
import io.flutter.plugin.common.MethodCall
import io.scanbot.genericdocument.entity.*
import io.scanbot.genericdocument.entity.FieldProperties.DisplayState
import io.scanbot.genericdocument.entity.GenericDocumentLibrary.wrap
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.barcode.entity.BarcodeDocumentFormat
import io.scanbot.sdk.barcode.entity.BarcodeFormat
import io.scanbot.sdk.barcode.entity.EngineMode
import io.scanbot.sdk.barcode.entity.MSIPlesseyChecksumAlgorithm
import io.scanbot.sdk.camera.CameraModule
import io.scanbot.sdk.flutter.utils.JsonUtils
import io.scanbot.sdk.flutter.utils.ObjectMapper
import io.scanbot.sdk.flutter.utils.StorageUtils
import io.scanbot.sdk.ui.camera.FinderAspectRatio
import io.scanbot.sdk.ui.result.ResultRepository
import io.scanbot.sdk.ui.result.ResultWrapper
import io.scanbot.sdk.ui.view.barcode.batch.configuration.BatchBarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerAdditionalConfiguration
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerConfiguration
import io.scanbot.sdk.ui.view.base.configuration.CameraOrientationMode
import io.scanbot.sdk.ui.view.base.configuration.OrientationMode
import io.scanbot.sdk.ui.view.camera.configuration.DocumentScannerConfiguration
import io.scanbot.sdk.ui.view.edit.configuration.CroppingConfiguration
import io.scanbot.sdk.ui.view.genericdocument.configuration.GenericDocumentRecognizerConfiguration
import io.scanbot.sdk.ui.view.hic.configuration.HealthInsuranceCardScannerConfiguration
import io.scanbot.sdk.ui.view.licenseplate.configuration.LicensePlateScannerConfiguration
import io.scanbot.sdk.ui.view.mrz.configuration.MRZScannerConfiguration
import io.scanbot.sdk.ui.view.nfc.configuration.NfcPassportConfiguration
import org.json.JSONException
import org.json.JSONObject
import java.util.*


fun readEhicScannerConfiguration(
    call: MethodCall,
    activity: Activity
): HealthInsuranceCardScannerConfiguration {
    val cameraConfiguration = HealthInsuranceCardScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

fun readLicensePlateScannerConfiguration(
    call: MethodCall,
    activity: Activity
): LicensePlateScannerConfiguration {
    val cameraConfiguration = LicensePlateScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}


fun readNFCPassportScannerConfiguration(
    call: MethodCall,
    activity: Activity
): NfcPassportConfiguration {
    val cameraConfiguration = NfcPassportConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

fun readBarcodeScannerConfiguration(
    call: MethodCall,
    activity: Activity
): BarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BarcodeScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        barcodeCameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)
    (arguments.get("additionalParameters") as? Json)?.let {
        barcodeCameraConfiguration.setAdditionalDetectionParameters(getAdditionalConfig(it))
    }
    if (arguments.containsKey("engineMode")) {
        val engineMode = arguments["engineMode"] as String
        val mode = EngineMode.valueOf(engineMode)
        barcodeCameraConfiguration.setEngineMode(mode)
    }
    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(
                BarcodeFormat.valueOf(
                    formatValue.replaceFirst(
                        "BarcodeFormat.",
                        ""
                    )
                )
            )
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }
    if (arguments.containsKey("acceptedDocumentFormats")) {
        val documentFormats = arguments["acceptedDocumentFormats"] as ArrayList<String>
        val nativeDocumentFormat = ArrayList<BarcodeDocumentFormat>()
        for (formatValue in documentFormats) {
            nativeDocumentFormat.add(
                BarcodeDocumentFormat.valueOf(
                    formatValue.replaceFirst(
                        "AcceptedDocumentFormats.",
                        ""
                    )
                )
            )
        }
        barcodeCameraConfiguration.setAcceptedDocumentFormats(nativeDocumentFormat)
    }
    if (arguments.containsKey("barcodeImageGenerationType")) {
        val imageGenType =
            BarcodeImageGenerationType.valueOf(arguments["barcodeImageGenerationType"] as String)
        barcodeCameraConfiguration.setBarcodeImageGenerationType(imageGenType)
    }
    return barcodeCameraConfiguration
}

private fun getAdditionalConfig(it: Json): BarcodeScannerAdditionalConfiguration {

    val defaultConfig = BarcodeScannerAdditionalConfiguration()
    val minimumTextLength = it["minimumTextLength"] as? Int
        ?: defaultConfig.minimumTextLength
    val enableGS1Decoding = it["enableGS1Decoding"] as? Boolean
        ?: defaultConfig.gs1DecodingEnabled
    val maximumTextLength = it["maximumTextLength"] as? Int
        ?: defaultConfig.maximumTextLength
    val minimum1DBarcodesQuietZone = it["minimum1DBarcodesQuietZone"] as? Int
        ?: defaultConfig.minimum1DQuietZoneSize
    val msiPlesseyChecksumAlgorithms = (it["msiPlesseyChecksumAlgorithm"] as? String)?.let { it1 ->
        getMsiPlesseyChecksumAlgorithmsConfig(
            it1
        )
    }
        ?: defaultConfig.msiPlesseyChecksumAlgorithms
    return BarcodeScannerAdditionalConfiguration(
        gs1DecodingEnabled = enableGS1Decoding,
        minimumTextLength = minimumTextLength,
        maximumTextLength = maximumTextLength,
        minimum1DQuietZoneSize = minimum1DBarcodesQuietZone,
        msiPlesseyChecksumAlgorithms = msiPlesseyChecksumAlgorithms
    )
}

private fun getMsiPlesseyChecksumAlgorithmsConfig(type: String): EnumSet<MSIPlesseyChecksumAlgorithm> {
    return when (type) {
        "None" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.None)
        "Mod10" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod10)
        "Mod1010" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1010)
        "Mod1110IBM" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1110IBM)
        "Mod1110NCR" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1110NCR)
        "Mod11IBM" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod11IBM)
        "Mod11NCR" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod11NCR)
        else -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod10)
    }
}

fun readBatchBarcodeScannerConfiguration(
    call: MethodCall,
    activity: Activity
): BatchBarcodeScannerConfiguration {
    val barcodeCameraConfiguration = BatchBarcodeScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        barcodeCameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        barcodeCameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, barcodeCameraConfiguration)
    (arguments.get("additionalParameters") as? Json)?.let {
        barcodeCameraConfiguration.setAdditionalDetectionParameters(getAdditionalConfig(it))
    }
    if (arguments.containsKey("engineMode")) {
        val engineMode = arguments["engineMode"] as String
        val mode = EngineMode.valueOf(engineMode)
        barcodeCameraConfiguration.setEngineMode(mode)
    }
    if (arguments.containsKey("barcodeFormats")) {
        val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
        val nativeBarcodeFormats = ArrayList<BarcodeFormat>()
        for (formatValue in barcodeFormats) {
            nativeBarcodeFormats.add(
                BarcodeFormat.valueOf(
                    formatValue.replaceFirst(
                        "BarcodeFormat.",
                        ""
                    )
                )
            )
        }
        barcodeCameraConfiguration.setBarcodeFormatsFilter(nativeBarcodeFormats)
    }
    if (arguments.containsKey("acceptedDocumentFormats")) {
        val documentFormats = arguments["acceptedDocumentFormats"] as ArrayList<String>
        val nativeDocumentFormat = ArrayList<BarcodeDocumentFormat>()
        for (formatValue in documentFormats) {
            nativeDocumentFormat.add(
                BarcodeDocumentFormat.valueOf(
                    formatValue.replaceFirst(
                        "AcceptedDocumentFormats.",
                        ""
                    )
                )
            )
        }
        barcodeCameraConfiguration.setAcceptedDocumentFormats(nativeDocumentFormat)
    }
    return barcodeCameraConfiguration
}

fun readCroppingConfiguration(
    call: MethodCall,
    activity: Activity,
    gson: Gson
): CroppingConfiguration {

    val arguments = call.arguments as Json
    val croppingConfiguration = call.argument<Json>("page")?.let {
        val page = it.toPage(gson = gson)
        CroppingConfiguration(page).apply {
            arguments.remove("page")
        }
    } ?: throw IllegalArgumentException("Page object is not set")

    val value = checkOrientationParam(call, arguments)
    value?.let {
        croppingConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, croppingConfiguration)
    return croppingConfiguration
}

fun readMRZScannerConfiguration(
    call: MethodCall,
    activity: Activity
): MRZScannerConfiguration {
    val cameraConfiguration = MRZScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    val aspectRatio = checkFinderAspectRatioParam(call, arguments)
    aspectRatio?.let {
        cameraConfiguration.setFinderAspectRatio(it)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}

private fun checkCameraOrientationParam(call: MethodCall, arguments: Json): CameraOrientationMode? {
    if (call.hasArgument("orientationLockMode")) {
        val value = when (call.argument<String>("orientationLockMode")) {
            "PORTRAIT" -> CameraOrientationMode.PORTRAIT
            "PORTRAIT_UPSIDE_DOWN" -> CameraOrientationMode.PORTRAIT
            "LANDSCAPE_LEFT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE_RIGHT" -> CameraOrientationMode.LANDSCAPE
            "LANDSCAPE" -> CameraOrientationMode.LANDSCAPE
            else -> CameraOrientationMode.PORTRAIT
        }
        arguments.remove("orientationLockMode")
        return value
    }
    return null
}

private fun checkOrientationParam(call: MethodCall, arguments: Json): OrientationMode? {
    if (call.hasArgument("orientationLockMode")) {
        val value = when (call.argument<String>("orientationLockMode")) {
            "PORTRAIT" -> OrientationMode.PORTRAIT
            "PORTRAIT_UPSIDE_DOWN" -> OrientationMode.PORTRAIT
            "LANDSCAPE_LEFT" -> OrientationMode.LANDSCAPE
            "LANDSCAPE_RIGHT" -> OrientationMode.LANDSCAPE
            "LANDSCAPE" -> OrientationMode.LANDSCAPE
            else -> OrientationMode.PORTRAIT
        }
        arguments.remove("orientationLockMode")
        return value
    }
    return null
}


private fun checkFinderAspectRatioParam(call: MethodCall, arguments: Json): FinderAspectRatio? {
    val propertyName = "finderAspectRatio"
    if (arguments[propertyName] != null) {
        val aspRatio = arguments[propertyName] as? Json
        val width = aspRatio?.get("width") as? Int
        val height = aspRatio?.get("height") as? Int
        arguments.remove(propertyName)
        return FinderAspectRatio(width?.toDouble() ?: 1.0, height?.toDouble() ?: 1.0)
    }
    return null
}

fun readDocumentScannerConfiguration(
    call: MethodCall,
    activity: Activity
): DocumentScannerConfiguration {
    val cameraConfiguration = DocumentScannerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        cameraConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, cameraConfiguration)
    return cameraConfiguration
}


fun readGenericDocumentRecognizerConfiguration(
    call: MethodCall,
    activity: Activity
): GenericDocumentRecognizerConfiguration {
    val genericDocumentRecognizerConfiguration = GenericDocumentRecognizerConfiguration()
    val arguments = call.arguments as Json
    val value = checkCameraOrientationParam(call, arguments)
    value?.let {
        genericDocumentRecognizerConfiguration.setOrientationLockMode(value)
    }
    ObjectMapper.map(arguments, genericDocumentRecognizerConfiguration)

    var acceptedDocumentTypes = ArrayList<RootDocumentType>()
    if (arguments.containsKey("acceptedDocumentTypes")) {
        val documentFormats = arguments["acceptedDocumentTypes"] as ArrayList<String>

        for (formatValue in documentFormats) {
            acceptedDocumentTypes.add(RootDocumentType.valueOf(formatValue))
        }
        genericDocumentRecognizerConfiguration.setAcceptedDocumentTypes(documentTypes = acceptedDocumentTypes)
    }

    //Todo: waiting for new android release. Talk to Ildar Regarding this
    if (arguments.containsKey("sharpnessAcceptanceFactor")) {
        var sharpnessAcceptanceFactor = arguments["sharpnessAcceptanceFactor"] as Int

        if (sharpnessAcceptanceFactor == 0) {
            sharpnessAcceptanceFactor = 1
        }
        genericDocumentRecognizerConfiguration.setSharpnessAcceptanceFactor(
            sharpnessAcceptanceFactor.toFloat()
        )
    }

    if (arguments.containsKey("cameraModule")) {
        var cameraModule = arguments["cameraModule"] as String
        if (cameraModule == "FRONT") {
            genericDocumentRecognizerConfiguration.setCameraModule(CameraModule.FRONT_MIRRORED)
        } else {
            genericDocumentRecognizerConfiguration.setCameraModule(CameraModule.valueOf(cameraModule))
        }
    }

    if (arguments.containsKey("detailsFieldConfiguration")) {
        if (acceptedDocumentTypes.isEmpty()) {
            acceptedDocumentTypes = ArrayList(RootDocumentType.ALL_TYPES)
        }
        val detailsFieldConfiguration = extractGenericDocumentRecognizerFieldProperties(
            arguments["detailsFieldConfiguration"] as Json,
            acceptedDocumentTypes
        )
        genericDocumentRecognizerConfiguration.setFieldsDisplayConfiguration(
            detailsFieldConfiguration
        )
    }

    // Inconsistency: on iOS 'submitButtonTitle' has been renamed to 'viewResultsButtonTitle'
    if (arguments.containsKey("viewResultsButtonTitle")) {
        val text = arguments["viewResultsButtonTitle"] as String
        if (text != null) {
            genericDocumentRecognizerConfiguration.setSubmitButtonTitle(text)
        }
    }

    return genericDocumentRecognizerConfiguration
}


private fun extractGenericDocumentRecognizerFieldProperties(
    detailsFieldConfiguration: Json,
    documentTypes: List<RootDocumentType>
): kotlin.collections.HashMap<String, FieldProperties> {
    val fields = HashMap<String, FieldProperties>()
    var configuration: JSONObject? = null
    try {
        configuration = JSONObject(detailsFieldConfiguration as Map<*, *>?)
    } catch (e: JSONException) {
        // ignore
    }

    if (configuration != null) {
        val deIdCardBackFields: HashMap<String?, String?> = object : HashMap<String?, String?>() {
            init {
                put(
                    "fieldAddressTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldAddressTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.ADDRESS
                )
                put(
                    "fieldEyeColorTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldEyeColorTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.EYE_COLOR
                )
                put(
                    "fieldHeightTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldHeightTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.HEIGHT
                )
                put(
                    "fieldIssueDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIssueDateTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.ISSUE_DATE
                )
                put(
                    "fieldIssuingAuthorityTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIssuingAuthorityTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.ISSUING_AUTHORITY
                )
                put(
                    "fieldRawMrzTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldRawMrzTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.RAW_MRZ
                )
                put(
                    "fieldPseudonymTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldPseudonymTitle")
                        ?: DeIdCardBack.NormalizedFieldNames.PSEUDONYM
                )
            }
        }
        val deIdCardFrontFields: HashMap<String?, String?> = object : HashMap<String?, String?>() {
            init {
                put(
                    "fieldBirthDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldBirthDateTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.BIRTH_DATE
                )
                put(
                    "fieldBirthPlaceTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldBirthPlaceTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.BIRTHPLACE
                )
                put(
                    "fieldExpiryDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldExpiryDateTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.EXPIRY_DATE
                )
                put(
                    "fieldGivenNamesTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldGivenNamesTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.GIVEN_NAMES
                )
                put(
                    "fieldIdTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIdTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.ID
                )
                put(
                    "fieldMaidenNameTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldMaidenNameTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.MAIDEN_NAME
                )
                put(
                    "fieldNationalityTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldNationalityTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.NATIONALITY
                )
                put(
                    "fieldPinTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldPinTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.PIN
                )
                put(
                    "fieldPhotoTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldPhotoTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.PHOTO
                )
                put(
                    "fieldSignatureTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldSignatureTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.SIGNATURE
                )
                put(
                    "fieldSurnameTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldSurnameTitle")
                        ?: DeIdCardFront.NormalizedFieldNames.SURNAME
                )
            }
        }
        val dePassportFields: HashMap<String?, String?> = object : HashMap<String?, String?>() {
            init {
                put(
                    "fieldBirthDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldBirthDateTitle")
                        ?: DePassport.NormalizedFieldNames.BIRTH_DATE
                )
                put(
                    "fieldBirthPlaceTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldBirthPlaceTitle")
                        ?: DePassport.NormalizedFieldNames.BIRTHPLACE
                )
                put(
                    "fieldCountryCodeTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldCountryCodeTitle")
                        ?: DePassport.NormalizedFieldNames.COUNTRY_CODE
                )
                put(
                    "fieldExpiryDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldExpiryDateTitle")
                        ?: DePassport.NormalizedFieldNames.EXPIRY_DATE
                )
                put(
                    "fieldGenderTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldGenderTitle")
                        ?: DePassport.NormalizedFieldNames.GENDER
                )
                put(
                    "fieldGivenNamesTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldGivenNamesTitle")
                        ?: DePassport.NormalizedFieldNames.GIVEN_NAMES
                )
                put(
                    "fieldIdTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIdTitle")
                        ?: DePassport.NormalizedFieldNames.ID
                )
                put(
                    "fieldIssueDateTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIssueDateTitle")
                        ?: DePassport.NormalizedFieldNames.ISSUE_DATE
                )
                put(
                    "fieldIssuingAuthorityTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldIssuingAuthorityTitle")
                        ?: DePassport.NormalizedFieldNames.ISSUING_AUTHORITY
                )
                put(
                    "fieldMaidenNameTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldMaidenNameTitle")
                        ?: DePassport.NormalizedFieldNames.MAIDEN_NAME
                )
                put(
                    "fieldNationalityTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldNationalityTitle")
                        ?: DePassport.NormalizedFieldNames.NATIONALITY
                )
                put(
                    "fieldPassportTypeTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldPassportTypeTitle")
                        ?: DePassport.NormalizedFieldNames.PASSPORT_TYPE
                )
                put(
                    "fieldPhotoTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldPhotoTitle")
                        ?: DePassport.NormalizedFieldNames.PHOTO
                )
                put(
                    "fieldRawMrzTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldRawMrzTitle")
                        ?: DePassport.NormalizedFieldNames.RAW_MRZ
                )
                put(
                    "fieldSignatureTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldSignatureTitle")
                        ?: DePassport.NormalizedFieldNames.SIGNATURE
                )
                put(
                    "fieldSurnameTitle",
                    JsonUtils.getStringOrNull(configuration, "fieldSurnameTitle")
                        ?: DePassport.NormalizedFieldNames.SURNAME
                )
            }
        }
        val deDriverLicenseBackFields: HashMap<String?, String?> =
            object : HashMap<String?, String?>() {
                init {
                    put(
                        "fieldRestrictionsTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldRestrictionsTitle")
                            ?: DeDriverLicenseBack.NormalizedFieldNames.RESTRICTIONS
                    )
                }
            }

        val deDriverLicenseFrontFields: HashMap<String?, String?> =
            object : HashMap<String?, String?>() {
                init {
                    put(
                        "fieldBirthDateTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldBirthDateTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.BIRTH_DATE
                    )
                    put(
                        "fieldBirthPlaceTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldBirthPlaceTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.BIRTHPLACE
                    )
                    put(
                        "fieldExpiryDateTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldExpiryDateTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.EXPIRY_DATE
                    )
                    put(
                        "fieldGivenNamesTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldGivenNamesTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.GIVEN_NAMES
                    )
                    put(
                        "fieldIdTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldIdTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.ID
                    )
                    put(
                        "fieldIssueDateTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldIssueDateTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.ISSUE_DATE
                    )
                    put(
                        "fieldIssuingAuthorityTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldIssuingAuthorityTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.ISSUING_AUTHORITY
                    )
                    put(
                        "fieldLicenseCategoriesTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldLicenseCategoriesTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.LICENSE_CATEGORIES
                    )
                    put(
                        "fieldPhotoTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldPhotoTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.PHOTO
                    )
                    put(
                        "fieldSignatureTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldSignatureTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.SIGNATURE
                    )
                    put(
                        "fieldSurnameTitle",
                        JsonUtils.getStringOrNull(configuration, "fieldSurnameTitle")
                            ?: DeDriverLicenseFront.NormalizedFieldNames.SURNAME
                    )

                }
            }
        if (documentTypes.contains(RootDocumentType.DeIdCardBack)) {
            for ((key, value) in deIdCardBackFields) {
                if (key != null && value != null) {
                    fields[key] = FieldProperties(value, DisplayState.AlwaysVisible)
                }
            }
        }
        if (documentTypes.contains(RootDocumentType.DeIdCardFront)) {
            for ((key, value) in deIdCardFrontFields) {
                if (key != null && value != null) {
                    fields[key] = FieldProperties(value, DisplayState.AlwaysVisible)
                }
            }
        }
        if (documentTypes.contains(RootDocumentType.DePassport)) {
            for ((key, value) in dePassportFields) {
                if (key != null && value != null) {
                    fields[key] = FieldProperties(value, DisplayState.AlwaysVisible)
                }
            }
        }
        if (documentTypes.contains(RootDocumentType.DeDriverLicenseBack)) {
            for ((key, value) in deDriverLicenseBackFields) {
                if (key != null && value != null) {
                    fields[key] = FieldProperties(value, DisplayState.AlwaysVisible)
                }
            }
        }
        if (documentTypes.contains(RootDocumentType.DeDriverLicenseFront)) {
            for ((key, value) in deDriverLicenseFrontFields) {
                if (key != null && value != null) {
                    fields[key] = FieldProperties(value, DisplayState.AlwaysVisible)
                }
            }
        }
    }
    return fields
}

private fun extractGenericDocumentRecognizerTitleProperties(
    detailsFieldConfiguration: Json,
    documentTypes: List<RootDocumentType>
): HashMap<String, DocumentProperties> {
    val properties = HashMap<String, DocumentProperties>()
    var configuration: JSONObject? = null
    try {
        configuration = JSONObject(detailsFieldConfiguration as Map<*, *>?)
    } catch (e: JSONException) {
        // ignore
    }
    if (configuration != null) {
        if (configuration.has("dePassportDocumentTitle") && documentTypes.contains(RootDocumentType.DePassport)) {
            val text = JsonUtils.getStringOrNull(configuration, "dePassportDocumentTitle")
            if (text != null) {
                properties[DePassport.DOCUMENT_NORMALIZED_TYPE] = DocumentProperties(text)
            }
        }
        if (configuration.has("deIdCardBackDocumentTitle") && documentTypes.contains(
                RootDocumentType.DeIdCardBack
            )
        ) {
            val text = JsonUtils.getStringOrNull(configuration, "deIdCardBackDocumentTitle")
            if (text != null) {
                properties[DeIdCardBack.DOCUMENT_NORMALIZED_TYPE] = DocumentProperties(text)
            }
        }
        if (configuration.has("deIdCardFrontDocumentTitle") && documentTypes.contains(
                RootDocumentType.DeIdCardFront
            )
        ) {
            val text = JsonUtils.getStringOrNull(configuration, "deIdCardFrontDocumentTitle")
            if (text != null) {
                properties[DeIdCardFront.DOCUMENT_NORMALIZED_TYPE] = DocumentProperties(text)
            }
        }
        if (configuration.has("deDriverLicenseBackDocumentTitle") && documentTypes.contains(
                RootDocumentType.DeDriverLicenseBack
            )
        ) {
            val text = JsonUtils.getStringOrNull(configuration, "deDriverLicenseBackDocumentTitle")
            if (text != null) {
                properties[DeDriverLicenseBack.DOCUMENT_NORMALIZED_TYPE] = DocumentProperties(text)
            }
        }
        if (configuration.has("deDriverLicenseFrontDocumentTitle") && documentTypes.contains(
                RootDocumentType.DeDriverLicenseFront
            )
        ) {
            val text = JsonUtils.getStringOrNull(configuration, "deDriverLicenseFrontDocumentTitle")
            if (text != null) {
                properties[DeDriverLicenseFront.DOCUMENT_NORMALIZED_TYPE] = DocumentProperties(text)
            }
        }
    }
    return properties
}

fun extractGenericDocumentRecognizerResultFromResultWrappers(
    resultWrappers: java.util.ArrayList<ResultWrapper<GenericDocument>>?,
    scanbotSDK: ScanbotSDK,
    context: Context
): JSONObject {
    var outArgs = JSONObject()
    if (resultWrappers == null) {
        return outArgs
    }
    var documentType: String? = ""
    val fields = JSONObject()
    val documents: List<GenericDocument> =
        extractDocumentsFromResultWrappers(resultWrappers, scanbotSDK)
    putGdrFieldFromDocuments(documents, fields, "surname", "Surname")
    putGdrFieldFromDocuments(documents, fields, "givenNames", "GivenNames")
    putGdrFieldFromDocuments(documents, fields, "nationality", "Nationality")
    putGdrFieldFromDocuments(documents, fields, "birthDate", "BirthDate")
    putGdrFieldFromDocuments(documents, fields, "rawMrz", "RawMRZ")
    putGdrFieldFromDocuments(documents, fields, "expiryDate", "ExpiryDate")
    putGdrFieldFromDocuments(documents, fields, "birthplace", "Birthplace")
    putGdrFieldFromDocuments(documents, fields, "issueDate", "IssueDate")
    putGdrFieldFromDocuments(documents, fields, "issuingAuthority", "IssuingAuthority")
    putGdrFieldFromDocuments(documents, fields, "address", "Address")
    putGdrFieldFromDocuments(documents, fields, "eyeColor", "EyeColor")
    putGdrFieldFromDocuments(documents, fields, "height", "Height")
    putGdrFieldFromDocuments(documents, fields, "pin", "PIN")
    putGdrFieldFromDocuments(documents, fields, "pseudonym", "Pseudonym")
    putGdrFieldFromDocuments(documents, fields, "countryCode", "CountryCode")
    putGdrFieldFromDocuments(documents, fields, "gender", "Gender")
    putGdrFieldFromDocuments(documents, fields, "maidenName", "MaidenName")
    putGdrFieldFromDocuments(documents, fields, "passportType", "PassportType")
    putGdrFieldFromDocuments(documents, fields, "id", "ID")
    putGdrFieldFromDocuments(documents, fields, "licenseCategories", "LicenseCategories")
    putGdrFieldFromDocuments(documents, fields, "restrictions", "Restrictions")
    putGdrFieldFromDocuments(documents, fields, "validFrom", "ValidFrom")
    putGdrFieldFromDocuments(documents, fields, "validUntil", "ValidUntil")
    putImageInGdrJsonMapFromDocuments(documents, fields, "photoImageUri", "Photo", context)
    putImageInGdrJsonMapFromDocuments(documents, fields, "signatureImageUri", "Signature", context)
    for (document in documents) {
        val wrap = document.wrap()
        if (wrap is DePassport) {
            val mrzFields = JSONObject()
            val mrz = wrap.mrz
            putGdrFieldFromTextFieldWrapper(mrzFields, "birthDate", mrz.birthDate)
            putGdrFieldFromTextFieldWrapper(mrzFields, "documentNumber", mrz.documentNumber)
            putGdrFieldFromTextFieldWrapper(mrzFields, "expiryDate", mrz.expiryDate)
            putGdrFieldFromTextFieldWrapper(mrzFields, "gender", mrz.gender)
            putGdrFieldFromTextFieldWrapper(mrzFields, "givenNames", mrz.givenNames)
            putGdrFieldFromTextFieldWrapper(mrzFields, "issuingAuthority", mrz.issuingAuthority)
            putGdrFieldFromTextFieldWrapper(mrzFields, "nationality", mrz.nationality)
            putGdrFieldFromTextFieldWrapper(mrzFields, "optional1", mrz.optional1)
            putGdrFieldFromTextFieldWrapper(mrzFields, "optional2", mrz.optional2)
            putGdrFieldFromTextFieldWrapper(mrzFields, "surname", mrz.surname)
            putGdrFieldFromTextFieldWrapper(mrzFields, "travelDocType", mrz.travelDocType)
            putGdrFieldFromTextFieldWrapper(
                mrzFields,
                "travelDocTypeVariant",
                mrz.travelDocTypeVariant
            )
            fields.put("mrz", mrzFields)
        } else if (wrap is DeDriverLicenseBack) {
            val driverLicenseCategories = JSONObject()
            val categories: DeDriverLicenseBack.Categories = wrap.categories
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "a", categories.a)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "a1", categories.a1)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "a2", categories.a2)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "b", categories.b)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "b1", categories.b1)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "be", categories.be)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "c", categories.c)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "c1", categories.c1)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "c1e", categories.c1e)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "ce", categories.ce)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "d", categories.d)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "d1", categories.d1)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "d1e", categories.d1e)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "de", categories.de)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "l", categories.l)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "m", categories.m)
            putGdrFieldFromDriverLicenseCategory(driverLicenseCategories, "t", categories.t)
            fields.put("categories", driverLicenseCategories)
        }

        documentType = getDocumentTypeFromGenericDocumentWrapper(wrap)
    }
    outArgs.put("documentType", documentType)
    outArgs.put("fields", fields)
    return outArgs
}

private fun extractDocumentsFromResultWrappers(
    wrappers: List<ResultWrapper<GenericDocument>>,
    scanbotSDK: ScanbotSDK
): List<GenericDocument> {
    val documents: MutableList<GenericDocument> = java.util.ArrayList()

    for ((resultId, clazz) in wrappers) {
        // Get the ResultRepository from the ScanbotSDK instance
        val resultRepository: ResultRepository<GenericDocument> =
            scanbotSDK.resultRepositoryForClass(clazz)

        // Receive an instance of GenericDocument class from the repository
        // This call will also remove the result from the repository (to make the memory usage less)
        val document = resultRepository.getResultAndErase(resultId) ?: continue
        documents.add(document)
    }
    return documents
}

private fun putGdrFieldFromTextFieldWrapper(
    dest: JSONObject,
    jsonKey: String,
    textFieldWrapper: TextFieldWrapper?
) {
    if (textFieldWrapper != null) {
        val ocrResult: JSONObject = getOcrResultFromOcrResult(textFieldWrapper.value)
        dest.put(jsonKey, ocrResult)
    }
}

private fun putGdrFieldFromDriverLicenseCategory(
    dest: JSONObject,
    jsonKey: String,
    category: DeDriverLicenseBack.Category?
) {
    if (category != null) {
        val categoryData = JSONObject()
        putGdrFieldFromTextFieldWrapper(categoryData, "restrictions", category.restrictions)
        putGdrFieldFromTextFieldWrapper(categoryData, "validFrom", category.validFrom)
        putGdrFieldFromTextFieldWrapper(categoryData, "validUntil", category.validUntil)
        dest.put(jsonKey, categoryData)
    }
}

private fun putGdrFieldFromDocuments(
    documents: List<GenericDocument>,
    dest: JSONObject,
    jsonKey: String,
    documentKey: String
) {
    for (document in documents) {
        val (_, value) = document.fieldByTypeName(documentKey) ?: continue
        val result = value ?: return
        val fieldMap: JSONObject = getOcrResultFromOcrResult(result)
        dest.put(jsonKey, fieldMap)
    }
}

private fun getOcrResultFromOcrResult(result: OcrResult): JSONObject {
    val fieldMap = JSONObject()
    fieldMap.put("text", result.text)
    fieldMap.put("confidence", result.confidence)
    return fieldMap
}

private fun putImageInGdrJsonMapFromDocuments(
    documents: List<GenericDocument>,
    dest: JSONObject,
    jsonKey: String,
    documentKey: String,
    context: Context
) {
    try {
        val photo = getImageFromGdrDocuments(documents, documentKey)
        if (photo != null) {
            val imageUri: Uri = StorageUtils.storeImage(photo, 100, context)
            dest.put(jsonKey, imageUri.toString())
        }
    } catch (e: Exception) {
        e.printStackTrace()
    }
}

private fun getImageFromGdrDocuments(
    documents: List<GenericDocument>,
    fieldKey: String
): Bitmap? {
    for (document in documents) {
        val (_, _, image) = document.fieldByTypeName(fieldKey) ?: continue
        return image ?: continue
    }
    return null
}


private fun getDocumentTypeFromGenericDocumentWrapper(wrap: GenericDocumentWrapper): String? {
    var documentType = ""
    if (wrap is DePassport) {
        documentType = "DePassport"
    } else if (wrap is DeDriverLicenseBack || wrap is DeDriverLicenseFront) {
        documentType = "DeDriverLicense"
    } else if (wrap is DeIdCardBack || wrap is DeIdCardFront) {
        documentType = "DeIdCard"
    }
    return documentType
}


