package io.scanbot.sdk.flutter.camera

import android.content.Context
import android.view.View
import android.widget.FrameLayout
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.platform.PlatformView
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.barcode.ui.BarcodeScannerView
import io.scanbot.sdk.flutter.ByteArrayToBase64TypeAdapter
import io.scanbot.sdk.flutter.Json
import io.scanbot.sdk.flutter.camera.DetectorType.*
import io.scanbot.sdk.flutter.camera.detectorWrappers.BarcodeDetector
import io.scanbot.sdk.flutter.camera.detectorWrappers.IDetectorWrapper
import io.scanbot.sdk.flutter.getLicenseStatusJson
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import kotlin.coroutines.CoroutineContext


class ScanbotCameraPlatformView internal constructor(
    private val context: Context,
    private val messenger: BinaryMessenger,
    private val id: Int,
    private val creationParams: Json?
) : PlatformView,
    MethodChannel.MethodCallHandler, CoroutineScope {

    private var rootView: FrameLayout = FrameLayout(context)
    private val scanbotSDK = ScanbotSDK(context)
    private val customViewDisplayOwner: CustomViewDisplayOwner = CustomViewDisplayOwner()
    private var cameraChannel = MethodChannel(messenger, "scanbot_sdk_camera_$id")
    private val detectorChannel = MethodChannel(messenger, "scanbot_sdk_camera.detector_$id")
    private val finderChannel = MethodChannel(
        messenger,
        "scanbot_sdk_camera.finder_${creationParams?.get("layout_orientation") ?: ""}"
    )
    private var currentDetector: IDetectorWrapper? = null
    private val returnContext = Dispatchers.Main

    val gson = GsonBuilder()
        .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
        .registerTypeAdapter(ByteArray::class.java, ByteArrayToBase64TypeAdapter())
        .create()

    private var job: Job = Job()
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.Default + job

    override fun getView(): View {
        return rootView
    }

    init {
        initCamera()
    }

    private fun initCamera() {
        cameraChannel.setMethodCallHandler(this)
        creationParams?.apply {
            checkParam<Json>("scannerConfiguration")?.let {
                attachDetector(it)
            }
        }
        customViewDisplayOwner.startListening()
    }

    override fun onMethodCall(methodCall: MethodCall, result: MethodChannel.Result) {
        when (methodCall.method) {
            "flashEnabled" -> {
                val useFlash = methodCall.argument<Boolean>("enabled") ?: false
                currentDetector?.flashEnabled(useFlash)
                result.success("success")
            }
            "toggleContinuousFocus" -> {
                currentDetector?.toggleContinuousFocus()
                result.success("success")
            }

            "pause" -> {
                customViewDisplayOwner.stopListening()
                result.success("success")
            }
            "resume" -> {
                customViewDisplayOwner.startListening()
                result.success("success")
            }
            "dispose" -> {
                customViewDisplayOwner.stopListening()
                dispose()
                result.success("success")
            }
            "isFlashAvailable" -> {
                result.success(currentDetector?.isFlashAvailable() ?: false)
            }
            else -> result.notImplemented()
        }
    }

    private fun attachDetector(arguments: Json) {
        clearPrevDetectorState()
        if (checkLicenseInvalid()) return
        val detectorType = valueOf((arguments["detectorType"] as? String) ?: "")
        currentDetector = when (detectorType) {
            barcode -> BarcodeScannerView(context).let {
                rootView.addView(it)
                BarcodeDetector(
                    it, this,
                    cameraChannel,
                    detectorChannel,
                    finderChannel,
                    scanbotSDK,
                    customViewDisplayOwner,
                    creationParams ?: HashMap(),
                    gson
                )
            }

            mrz -> TODO("Mrz scanner is not implemented yet")
            ehic -> TODO("Health insurance scanner is not implemented yet")
            payform -> TODO("Payform scanner is not implemented yet")
            document -> TODO("Document scanner is not implemented yet")
        }

        currentDetector?.attach(arguments)
    }

    private fun checkLicenseInvalid(): Boolean {
        if (!scanbotSDK.licenseInfo.isValid) {
            launch(returnContext) {
                val json = scanbotSDK.getLicenseStatusJson()
                detectorChannel.invokeMethod(
                    "onLicenseError",
                    json.toString()
                )
            }
            return true
        }
        return false
    }

    private fun clearPrevDetectorState() {
        rootView.removeAllViews()
        currentDetector?.detach()
        currentDetector = null
    }


    override fun dispose() {
        clearPrevDetectorState()
        job.cancel()
        customViewDisplayOwner.stopListening()
    }

    override fun onFlutterViewDetached() {
        super.onFlutterViewDetached()
    }

    override fun onFlutterViewAttached(flutterView: View) {
        super.onFlutterViewAttached(flutterView)
    }

}

enum class DetectorType {
    barcode,
    mrz,
    ehic,
    document,
    payform,
}

fun <T> Map<String, Any>.checkParam(name: String, value: (T) -> Unit = {}): T? {
    return this.get(name)?.also {
        value(it as T)
    } as? T
}
