package io.scanbot.sdk.flutter.camera.detectorWrappers

import android.content.pm.PackageManager
import android.graphics.Color
import android.util.Log
import android.view.View.OnLayoutChangeListener
import android.widget.FrameLayout
import com.google.gson.Gson
import com.google.gson.JsonObject
import io.flutter.plugin.common.MethodChannel
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.barcode.BarcodeDetectorFrameHandler
import io.scanbot.sdk.barcode.ScanbotBarcodeDetector
import io.scanbot.sdk.barcode.entity.BarcodeScanningResult
import io.scanbot.sdk.barcode.ui.BarcodeScannerView
import io.scanbot.sdk.barcode.ui.IBarcodeScannerViewCallback
import io.scanbot.sdk.camera.CaptureInfo
import io.scanbot.sdk.camera.FrameHandlerResult
import io.scanbot.sdk.flutter.*
import io.scanbot.sdk.flutter.camera.CustomViewDisplayOwner
import io.scanbot.sdk.flutter.camera.checkParam
import io.scanbot.sdk.flutter.camera.config.BarcodeScannerConfiguration
import io.scanbot.sdk.flutter.camera.usecase.SaveBarcodePreviewFrameUseCase
import io.scanbot.sdk.flutter.camera.usecase.SaveBarcodeTakenPictureUseCase
import io.scanbot.sdk.flutter.utils.convertDpToPx
import io.scanbot.sdk.flutter.utils.convertPxToDp
import io.scanbot.sdk.ui.camera.CameraUiSettings
import io.scanbot.sdk.ui.camera.ScanbotCameraContainerView
import io.scanbot.sdk.ui.camera.ScanbotCameraXView
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeImageGenerationType.*
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.single
import java.util.*


class BarcodeDetector(
    private val scannerView: BarcodeScannerView,
    private val coroutineScope: CoroutineScope,
    private val cameraChannel: MethodChannel,
    private val detectorChannel: MethodChannel,
    private val finderChannel: MethodChannel,
    override val scanbotSDK: ScanbotSDK,
    private val customViewDisplayOwner: CustomViewDisplayOwner,
    private val creationParams: HashMap<String, Any>,
    val gson: Gson
) : IDetectorWrapper, CoroutineScope by coroutineScope {
    private val returnContext = Dispatchers.Main
    private val barcodeFileStorage = scanbotSDK.createBarcodeFileStorage()
    private val imageFileIOProcessor = scanbotSDK.fileIOProcessor()

    private val saveBarcodePreviewFrameUseCase =
        SaveBarcodePreviewFrameUseCase(barcodeFileStorage, imageFileIOProcessor)
    private val saveBarcodeTakenPictureUseCase =
        SaveBarcodeTakenPictureUseCase(barcodeFileStorage, imageFileIOProcessor)

    private var cachedBarcodeResult: BarcodeScanningResult? = null
    private var barcodeImageGenerationType: BarcodeImageGenerationType? = null
    private var isSavingBarcodeSnapshot = false

    private val handler = BarcodeDetectorFrameHandler.ResultHandler { result ->
        runBlocking {
            when (result) {
                is FrameHandlerResult.Success -> {
                    result.value?.let { result ->
                        try {
                            val json = gson.toJsonTree(result).asJsonObject
                            when (barcodeImageGenerationType) {
                                NONE -> {
                                    withContext(returnContext) {
                                        detectorChannel.invokeMethod(
                                            "onBarcodeDetected",
                                            gson.toJson(json)
                                        )
                                    }
                                }
                                VIDEO_FRAME -> {
                                    savePreviewFrameAndReturn(result, json)
                                    withContext(returnContext) {
                                        detectorChannel.invokeMethod(
                                            "onBarcodeDetected",
                                            gson.toJson(json)
                                        )
                                    }
                                }
                                CAPTURED_IMAGE -> {
                                    if (!isSavingBarcodeSnapshot) {
                                        cachedBarcodeResult = result
                                    }
                                }
                            }
                        } catch (e: Throwable) {
                            e.printStackTrace()
                        }
                    } ?: run { cachedBarcodeResult == null }
                }
                is FrameHandlerResult.Failure -> {
                    withContext(returnContext) {
                        val json = scanbotSDK.getLicenseStatusJson()
                        detectorChannel.invokeMethod(
                            "onLicenseError",
                            json.toString()
                        )
                    }
                }
            }
        }
        false
    }

    override fun attach(arguments: Json) {
        creationParams.apply {
            scannerView.initCamera(CameraUiSettings(true))
            val cameraXView =
                scannerView.findViewById<ScanbotCameraContainerView>(R.id.scanbot_camera_view)
            (cameraXView.getChildAt(0) as ScanbotCameraXView).lifecycleOwner =
                customViewDisplayOwner
            configureFinderView()
            val barcodeScannerConfiguration = (arguments).toBarcodeScannerConfig(gson)
            val barcodeDetector = scanbotSDK.createBarcodeDetector()

            barcodeScannerConfiguration.detectionIntervalInMs?.let {
                scannerView.viewController.barcodeDetectionInterval = it.toLong()
            }

            updateScannerConfiguration(barcodeDetector, barcodeScannerConfiguration)

            addDetectorMethodHandler()
            scannerView.initDetectionBehavior(
                barcodeDetector,
                handler,
                object : IBarcodeScannerViewCallback {
                    override fun onCameraOpen() {
                        creationParams.apply {
                            launch {
                                withContext(returnContext) {
                                    checkParam<Boolean>("flashEnabled")?.let {
                                        scannerView.viewController.useFlash(it)
                                    }
                                    checkParam<Float>("cameraZoomFactor")?.let {
                                        cameraXView.setOpticalZoomLevel(it)
                                    }

                                    cameraChannel.invokeMethod("onCameraOpen", "")
                                }
                            }
                        }
                    }

                    override fun onPictureTaken(image: ByteArray, captureInfo: CaptureInfo) {
                        //todo possibly need to fix threading
                        launch {
                            saveSnappedImageAndReturn(image, captureInfo)
                            isSavingBarcodeSnapshot = false
                        }
                    }
                })

            barcodeImageGenerationType = barcodeScannerConfiguration.barcodeImageGenerationType
            when (barcodeScannerConfiguration.barcodeImageGenerationType) {
                VIDEO_FRAME -> barcodeDetector.modifyConfig { this.setSaveCameraPreviewFrame(true) }
                CAPTURED_IMAGE -> {
                    scannerView.viewController.autoSnappingEnabled = true
                    barcodeScannerConfiguration.autoSnappingSensitivity?.let {
                        scannerView.viewController.setAutoSnappingSensitivity(
                            it
                        )
                    }
                }
                else -> {
                }
            }
        }
    }

    private suspend fun saveSnappedImageAndReturn(image: ByteArray, captureInfo: CaptureInfo) {
        try {
            val json = cachedBarcodeResult?.let { gson.toJsonTree(it).asJsonObject } ?: JsonObject()
            withContext(returnContext) { cameraChannel.invokeMethod("onShowHeavyProcessing", true) }
            val filePath =
                saveBarcodeTakenPictureUseCase.saveTakenPicture(image, captureInfo.imageOrientation)
                    .single()
            json.addProperty("barcodeImageURI", filePath)
            withContext(returnContext) {
                cameraChannel.invokeMethod(
                    "onShowHeavyProcessing",
                    false
                )
            }
            withContext(returnContext) {
                detectorChannel.invokeMethod(
                    "onBarcodeDetected",
                    gson.toJson(json)
                )
            }
        } catch (e: Throwable) {
            Log.e("ScanbotCameraModule", "error while saving snapped image", e)
        }
    }

    private suspend fun savePreviewFrameAndReturn(result: BarcodeScanningResult, json: JsonObject) {
        result.previewFrame?.let {
            withContext(returnContext) { cameraChannel.invokeMethod("onShowHeavyProcessing", true) }
            val filePath = saveBarcodePreviewFrameUseCase.savePreviewFrame(it).single()
            json.addProperty("barcodeImageURI", filePath)
            withContext(returnContext) {
                cameraChannel.invokeMethod(
                    "onShowHeavyProcessing",
                    false
                )
            }
        }
    }

    private fun updateScannerConfiguration(
        barcodeDetector: ScanbotBarcodeDetector,
        barcodeScannerConfiguration: BarcodeScannerConfiguration
    ) {
        barcodeDetector.modifyConfig {
            barcodeScannerConfiguration.acceptedDocumentFormats?.let {
                this.setAcceptedDocumentFormats(it)
            }
            barcodeScannerConfiguration.barcodeFormats?.let { this.setBarcodeFormats(it) }
            barcodeScannerConfiguration.engineMode?.let { this.setEngineMode(it) }
            barcodeScannerConfiguration.additionalParameters?.let { params ->
                this.modifyAdditionalConfig {
                    params.enableGS1Decoding?.let { this.setGs1DecodingEnabled(it) }
                    params.maximumTextLength?.let { this.setMaximumTextLength(it) }
                    params.stripCheckDigits?.let { this.setStripCheckDigits(it) }
                    params.minimumTextLength?.let { this.setMinimumTextLength(it) }

                    params.minimum1DBarcodesQuietZone?.let { this.setMinimum1DQuietZoneSize(it) }
                    params.msiPlesseyChecksumAlgorithm?.let {
                        this.setMsiPlesseyChecksumAlgorithms(
                            EnumSet.of(it)
                        )
                    }
                }
            }
        }
    }

    private var frameLayout: FrameLayout? = null

    private val finderListener =
        OnLayoutChangeListener { view, left, top, right, bottom, oldLeft, oldTop, oldRight, oldBottom ->
            frameLayout?.notifyFlutterFinder(
                left, top, right, bottom
            )
        }

    private fun FrameLayout.registerFinderListener() {
        frameLayout = this
        addOnLayoutChangeListener(finderListener)
    }

    private fun FrameLayout.notifyFlutterFinder(
        left: Int,
        top: Int,
        right: Int,
        bottom: Int
    ) {
        finderChannel.invokeMethod(
            "finderSizeChanged", mapOf(
                "left" to convertPxToDp(context, left.toFloat()),
                "top" to convertPxToDp(context, top.toFloat()),
                "right" to convertPxToDp(context, right.toFloat()),
                "bottom" to convertPxToDp(context, bottom.toFloat()),
            )
        )
    }

    private fun Json.configureFinderView() {
        val finderViewController = scannerView.finderViewController
        checkParam<Json>("finder") {
            it.toFinderConfig(gson).also { config ->
                finderViewController.setStrokeColor(Color.TRANSPARENT)
                finderViewController.setOverlayColor(Color.TRANSPARENT)
                config.finderAspectRatio?.let {
                    finderViewController.setRequiredAspectRatios(listOf(it))
                }
                config.finderVerticalOffset?.let {
                    finderViewController.verticalOffset =
                        (convertDpToPx(scannerView.context, it).toInt())
                }
                config.finderMinimumPadding?.let {
                    finderViewController.setFinderMinPadding(
                        convertDpToPx(scannerView.context, it).toInt()
                    )
                }

                finderViewController.addFinderPlaceholder(
                    FrameLayout(scannerView.context).apply {
                        registerFinderListener()
                    })
            }
            finderViewController.setFinderEnabled(true)
        } ?: finderViewController.setFinderEnabled(false)
    }

    private fun addDetectorMethodHandler() {
        detectorChannel.setMethodCallHandler { call, result ->
            when (call.method) {
                "disable" -> {
                    scannerView.viewController.isFrameProcessingEnabled = false
                }
                "enable" -> {
                    scannerView.viewController.isFrameProcessingEnabled = true
                }
            }
        }
    }

    override fun detach() {
        frameLayout?.removeOnLayoutChangeListener(finderListener)
        detectorChannel.setMethodCallHandler(null)
    }

    override fun toggleContinuousFocus() {
        scannerView.viewController.continuousFocus()
    }

    override fun isFlashAvailable(): Boolean {
      return scannerView.context.packageManager.hasSystemFeature(PackageManager.FEATURE_CAMERA_FLASH)
    }

    override fun flashEnabled(enabled: Boolean) {
        scannerView.viewController.useFlash(enabled)
    }
}
