//
//  Renderer.swift
//  Pods-Runner
//
//  Created by Zhenya Knizhnik on 02.09.19.
//

import Foundation
import ScanbotSDK
import Flutter

final class Renderer {
    func renderPDF(from pages: [SBSDKUIPage],
                   pageSize: SBSDKPDFRendererPageSize = .fromImage,
                   result: @escaping FlutterResult) {
        if pages.isEmpty {
            result(FlutterError(code: "1010", message: "ERROR_iOS: No pages passed", details: "Rendering PDF error"))
            return
        }
        DispatchQueue.global(qos: .background).async {
            let storage = SBSDKIndexedImageStorage.temporaryStorage(for: pages)
            let outputURL = PluginPreferences.shared.newStorageCustomFileURL(extensionString: "pdf")
            
            // Try to change to SBSDKPDFUIRenderer after update
            SBSDKPDFRenderer.renderImageStorage(storage,
                                                copyImageStorage: false,
                                                indexSet: nil,
                                                with: pageSize,
                                                encrypter: ScanbotSDKUI.defaultPDFEncrypter(),
                                                output: outputURL,
                                                completionHandler: { (finished, error, info) in
                self.pdfProcessCompletion(finished: finished, error: error, info: info, result: result)
                storage.removeAllImages()
            })
        }
    }
    
    private func pdfProcessCompletion(finished: Bool,
                                      error: Error?,
                                      info: [String: NSObject]?,
                                      result: @escaping FlutterResult) {
        DispatchQueue.main.async {
            if let error = error {
                result(FlutterError(code: "1006", message: "ERROR_iOS: " + error.localizedDescription, details: "Rendering PDF error"))
            } else if finished, let outputURL = info?[SBSDKResultInfoDestinationFileURLKey] as? URL {
                result(outputURL.absoluteString)
            } else {
                result(FlutterError(code: "1007", message: "ERROR_iOS: Unknown rendering error", details: "Rendering PDF error"))
            }
        }
    }
    
    func renderTIFF(from pages: [SBSDKUIPage],
                    options: TIFFOptions,
                    result: @escaping FlutterResult) {
        
        if pages.isEmpty {
            result(FlutterError(code: "1011", message: "ERROR_iOS: No pages passed", details: "Rendering TIFF error"))
            return
        }
        DispatchQueue.global(qos: .background).async {
            let outputURL = PluginPreferences.shared.newStorageCustomFileURL(extensionString: "tiff")
            let imageURLs = pages.compactMap({ $0.documentImageURL() })
            
            let params = SBSDKTIFFImageWriterParameters.default()

            params.dpi = options.dpi
            params.binarize = options.isBinarized
            params.compression = options.compression

            var success: Bool = false
            // If we have an encrypter, we need to decrypt images before processing.
            if let pdfEncrypter = ScanbotSDKUI.defaultPDFEncrypter() {
                //We are using PDF encrypter, as it should be the same decrypter for everything.
                //We are mimicking android SDK behavior this way.
                let decryptedImages = self.decryptedImages(from: imageURLs)
                success = SBSDKTIFFImageWriter.writeTIFF(decryptedImages,
                                                         fileURL: outputURL,
                                                         encrypter: pdfEncrypter,
                                                         parameters: params)
            } else {
                success = SBSDKTIFFImageWriter.writeTIFF(from: imageURLs,
                                                         fileURL: outputURL,
                                                         parameters: params)
            }

            
            DispatchQueue.main.async {
                if success {
                    result(outputURL.absoluteString)
                } else {
                    result(FlutterError(code: "1008", message: "ERROR_iOS: Unknown rendering error", details: "Rendering TIFF error"))
                }
            }
        }
    }
    
    private func decryptedImages(from urls: [URL]) -> [UIImage] {
        var resultedImages: [UIImage] = []
        let encrypter = ScanbotSDKUI.defaultPDFEncrypter()
        for url in urls {
            if let data = try? Data(contentsOf: url),
               let decrytedData = encrypter?.decryptData(data),
               let image = UIImage(data: decrytedData) {
                
                resultedImages.append(image)
            }
        }
        return resultedImages
    }
}
