//
//  RTUUIPresenter.swift
//  scanbot_sdk
//
//  Created by Yevgeniy Knizhnik on 15.06.21.
//

import Foundation
import Flutter
import ScanbotSDK

class RTUUIPresenter: CallErrorHandler {
    
    private lazy var documentScannerPresenter = DocumentScannerPresenter()
    func showDocumentScanner(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIDocumentScannerConfiguration(from: arguments)
        
        if let presenter = presenter {
            self.documentScannerPresenter.present(on: presenter, with: config, result: parameters.result)
        }
    }
    
    private lazy var barcodeScannerPresenter = BarcodeScannerPresenter()
    func showBarcodeScanner(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIBarcodeScannerConfiguration(from: arguments)
        let codes = SBSDKBarcodeType.fromArguments(arguments)
        let documentTypes = SBSDKBarcodeDocumentType.fromArguments(arguments)
        
        guard let presenter = presenter else {
            return
        }
        
        self.barcodeScannerPresenter.present(on: presenter,
                                             with: config,
                                             machineCodeTypes: codes,
                                             barcodeDocumentTypes: documentTypes,
                                             result: result)
    }
    
    private lazy var mrzScannerPresenter = MRZScannerPresenter()
    func showMRZScanner(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIMRZScannerConfiguration(from: arguments)
        
        if let presenter = presenter {
            self.mrzScannerPresenter.present(on: presenter, with: config, result: result)
        }
    }
    
    private lazy var ehicScannerPresenter = EHICRecognizerPresenter()
    func showEHICScanner(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIHealthInsuranceCardScannerConfiguration(from: arguments)
        
        if let presenter = presenter {
            self.ehicScannerPresenter.present(on: presenter, with: config, result: result)
        }
    }
    
    @available(iOS 13.0, *)
    private lazy var nfcReaderPresenter = NFCReaderPresenter()
    func showNFCReader(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUINFCPassportReaderConfiguration(from: arguments)
        
        if let presenter = presenter {
            if #available(iOS 13.0, *) {
                self.nfcReaderPresenter.present(on: presenter, with: config, result: result)
            } else {
                self.returnCustomError(call: call,
                                       message: "NFC passport reader is available only for iOS 13 and higher",
                                       result: result)
            }
        }
    }
    
    private lazy var cropScreenPresenter = CropScreenPresenter()
    func showCropScreen(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any],
              let page = SBSDKUIPage(from: arguments) else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUICroppingScreenConfiguration(from: arguments)
        
        if let presenter = presenter {
            self.cropScreenPresenter.present(on: presenter, with: page, configuration: config, result: result)
        }
    }
    
    private lazy var batchScannerPresenter = BatchScannerPresenter()
    func showBatchScanner(_ parameters: CallUIParameters, dataFormatterChannel: FlutterMethodChannel?) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIBarcodesBatchScannerConfiguration(from: arguments)
        let codes = SBSDKBarcodeType.fromArguments(arguments)
        let documentTypes = SBSDKBarcodeDocumentType.fromArguments(arguments)
        
        if let presenter = presenter {
            var dataMapper: SBSDKUIBarcodesBatchScannerMapping?
            if let shouldUseDataMapper = arguments["shouldUseBarcodeFormatter"] as? Bool, shouldUseDataMapper == true {
                dataMapper = BarcodeBatchDataMapper(dataChannel: dataFormatterChannel)
            }
            
            self.batchScannerPresenter.present(on: presenter,
                                               with: config,
                                               dataMapper: dataMapper,
                                               barcodeTypes: codes,
                                               barcodeDocumentTypes: documentTypes,
                                               result: result)
        }
    }
    
    private lazy var licensePlateScannerPresenter = LicensePlateScannerPresenter()
    func showLicensePlateScanner(_ parameters: CallUIParameters) {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUILicensePlateScannerConfiguration(from: arguments)
        
        if let presenter = presenter {
            self.licensePlateScannerPresenter.present(on: presenter, with: config, result: result)
        }
    }
    
    private lazy var genericDocumentRecognizerPresenter = GenericDocumentRecognizerPresenter()
    func showGenericDocumentRecognizer(_ parameters: CallUIParameters)  {
        let (call, result, presenter) = (parameters.call, parameters.result, parameters.presenter)
        
        guard let arguments = call.arguments as? [String: Any] else {
            self.returnArgumentsError(call: call, result: result)
            return
        }
        let config = SBSDKUIGenericDocumentRecognizerConfiguration(from: arguments)
        
        if let presenter = presenter {
            
            self.genericDocumentRecognizerPresenter.present(on: presenter, with: config, result: parameters.result)
        }
    }
}
