import 'package:json_annotation/json_annotation.dart';
import 'package:scanbot_sdk/barcode_parsers/barcode_scanner.dart';

part 'aamva.g.dart';

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class AAMVADocument extends BarcodeFormattedResult {
  static const String DOCUMENT_FORMAT = 'AAMVA';

  /// Header raw string
  String headerRawString;

  /// AAMVA file type
  String fileType;

  /// Issuer identification number
  String issuerIdentificationNumber;

  /// Document version string
  String AAMVAVersionNumber;

  /// Jurisdiction version string
  String jurisdictionVersionNumber;

  /// Number of entries in the documents
  int numberOfEntries;

  /// Vector of document's subfiles
  List<AAMVADocumentSubfile> subfiles;

  AAMVADocument(
      this.headerRawString,
      this.fileType,
      this.issuerIdentificationNumber,
      this.AAMVAVersionNumber,
      this.jurisdictionVersionNumber,
      this.numberOfEntries,
      this.subfiles)
      : super(DOCUMENT_FORMAT);

  factory AAMVADocument.fromJson(Map<String, dynamic> json) =>
      _$AAMVADocumentFromJson(json);

  @override
  Map<String, dynamic> toJson() => _$AAMVADocumentToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)

/// AAMVA document subfile
class AAMVADocumentSubfile {
  /// Type of this subfile
  String subFileType;

  /// Subfile raw header string
  String subFileRawHeader;

  /// Parsed fields
  List<AAMVADocumentSubfileField> fields;

  AAMVADocumentSubfile(this.subFileType, this.subFileRawHeader, this.fields);

  factory AAMVADocumentSubfile.fromJson(Map<String, dynamic> json) =>
      _$AAMVADocumentSubfileFromJson(json);

  Map<String, dynamic> toJson() => _$AAMVADocumentSubfileToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)

/// AAMVA document subfile field
class AAMVADocumentSubfileField {
  /// Field type */
  AAMVARecordFieldType type;

  /// Field type raw string
  String typeString;

  /// Field formatted value
  String value;

  /// Field raw value
  String rawValue;

  AAMVADocumentSubfileField(
      this.type, this.typeString, this.value, this.rawValue);

  factory AAMVADocumentSubfileField.fromJson(Map<String, dynamic> json) =>
      _$AAMVADocumentSubfileFieldFromJson(json);

  Map<String, dynamic> toJson() => _$AAMVADocumentSubfileFieldToJson(this);
}

/// AAMVA subfile field type
enum AAMVARecordFieldType {
  Unknown,
  FullName, //DAA, DBN
  LastName, //DAB, DBO, DCS
  FamilyName, //DAB, DCS, DBO
  FirstName, //DAC, DCT, DBP
  MiddleNameOrInitial, //DAD, DBQ
  NameSuffix, //DAE
  NamePrefix, //DAF
  MailingStreetAddress1, //DAG
  MailingStreetAddress2, //DAH
  MailingCity, //DAI
  MailingJurisdictionCode, //DAJ
  MailingPostalCode, //DAK
  ResidenceStreetAddress1, //DAL
  ResidenceStreetAddress2, //DAM
  ResidenceCity, //DAN
  ResidenceJurisdictionCode, //DAO
  ResidencePostalCode, //DAP
  LicenseOrIdNumber, //DAQ
  LicenseClassificationCode, //DAR
  LicenseRestrictionCode, //DAS
  LicenseEndorsementsCode, //DAT
  HeightInFeetInches, //DAU
  HeightInCM, //DAV
  WeightInLBS, //DAW
  WeightInKG, //DAX
  EyesColor, //DAY
  HairColor, //DAZ
  LicenseExpirationDate, //DBA
  DateOfBirth, //DBB, DBL
  Sex, //DBC
  LicenseOrIdDocumentIssueDate, //DBD
  IssueTimestamp, //DBE
  NumberOfDuplicates, //DBF
  MedicalIndicatorCodes, //DBG
  OrganDonor, //DBH
  NonResidentIndicator, //DBI
  UniqueCustomerIdentifier, //DBJ
  SocialSecurityNumber, //DBK, DBM
  Suffix, //DBR, DCU
  Prefix, //DBS
  VirginiaSpecificClass, //DCA
  VirginiaSpecificRestrictions, //DCB
  VirginiaSpecificEndorsements, //DCD
  PhysicalDescriptionWeightRange, //DCE
  DocumentDiscriminator, //DCF
  CountryTerritoryOfIssuance, //DCG
  FederalCommercialVehicleCodes, //DCH
  PlaceOfBirth, //DCI
  AuditInformation, //DCJ
  InventoryControlNumber, //DCK
  RaceEthnicity, //DCL
  StandardVehicleClassification, //DCM
  StandardEndorsementCode, //DCN
  StandardRestrictionCode, //DCO
  JurisdictionSpecificVehicleClassificationDescription, //DCP
  JurisdictionSpecific, //DCQ
  JurisdictionSpecificRestrictionCodeDescription, //DCR
  ComplianceType, //DDA
  CardRevisionDate, //DDB
  HazMatEndorsementExpiryDate, //DDC
  LimitedDurationDocumentIndicator, //DDD
  FamilyNameTruncation, //DDE
  FirstNamesTruncation, //DDF
  MiddleNamesTruncation, //DDG
  Under18Until, //DDH
  Under19Until, //DDI
  Under21Until, //DDJ
  OrganDonorIndicator, //DDK
  VeteranIndicator, //DDL
  PermitClassificationCode, //PAA
  PermitExpirationDate, //PAB
  PermitIdentifier, //PAC
  PermitIssueDate, //PAD
  PermitRestrictionCode, //PAE
  PermitEndorsementCode, //PAF
  CourtRestrictionCode, //ZVA
}
