import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/cupertino.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:scanbot_sdk/barcode_parsers/barcode_scanner.dart';
import 'package:scanbot_sdk/common_data.dart';
import 'package:scanbot_sdk/sdk_utils.dart';

import 'common_data.dart';

part 'barcode_scanning_data.g.dart';

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeScanningResult extends Result {
  List<BarcodeItem> barcodeItems = <BarcodeItem>[];

  Uri? barcodeImageURI;

  BarcodeScanningResult(this.barcodeItems);

  /// We can use this method/'named contructor' while detection of Multiple images
  BarcodeScanningResult.fromUrlAndCode(this.barcodeImageURI, this.barcodeItems);

  BarcodeScanningResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  BarcodeScanningResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory BarcodeScanningResult.fromJson(Map<String, dynamic> json) =>
      _$BarcodeScanningResultFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeScanningResultToJson(this);
}

///Class created for handling the multiple images barcodes detection
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class MultiImageBarcodeScanningResult extends Result {
  List<BarcodeScanningResult> barcodeResults = <BarcodeScanningResult>[];

  MultiImageBarcodeScanningResult(this.barcodeResults);

  MultiImageBarcodeScanningResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
    if (barcodeResults == null) {
      barcodeResults = new List.empty();
    }
  }

  MultiImageBarcodeScanningResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory MultiImageBarcodeScanningResult.fromJson(Map<String, dynamic> json) =>
      _$MultiImageBarcodeScanningResultFromJson(json);

  Map<String, dynamic> toJson() =>
      _$MultiImageBarcodeScanningResultToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeItem {
  BarcodeFormattedResult? formattedResult;

  ///Scanned barcode format
  BarcodeFormat? barcodeFormat;

  ///Result data from barcode
  String? text;

  @JsonKey(fromJson: uintListFromString)

  ///raw data of barcode
  Uint8List? rawBytes;

  BarcodeItem(this.barcodeFormat, this.text);

  factory BarcodeItem.fromJson(Map<String, dynamic> json) =>
      _$BarcodeItemFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeItemToJson(this);
}

///Available barcode formats
enum BarcodeFormat {
  AZTEC,
  CODABAR,
  CODE_39,
  CODE_93,
  CODE_128,
  DATA_MATRIX,
  EAN_8,
  EAN_13,
  ITF,
  PDF_417,
  QR_CODE,
  RSS_14,
  RSS_EXPANDED,
  UPC_A,
  UPC_E,
  MSI_PLESSEY,
  UNKNOWN
}

///Available barcode document formats
enum AcceptedDocumentFormats {
  AAMVA,
  BOARDING_PASS,
  DE_MEDICAL_PLAN,
  MEDICAL_CERTIFICATE,
  ID_CARD_PDF_417,
  SEPA,
  SWISS_QR,
  VCARD
}

///Available engine modes for barcode V3 scanner version
enum EngineMode { Legacy, NextGen }

class PredefinedBarcodes {
  ///return list of one dimensional barcode formats
  static List<BarcodeFormat> allBarcodeTypes() {
    return [
      BarcodeFormat.AZTEC,
      BarcodeFormat.CODABAR,
      BarcodeFormat.CODE_39,
      BarcodeFormat.CODE_93,
      BarcodeFormat.CODE_128,
      BarcodeFormat.DATA_MATRIX,
      BarcodeFormat.EAN_8,
      BarcodeFormat.EAN_13,
      BarcodeFormat.ITF,
      BarcodeFormat.PDF_417,
      BarcodeFormat.QR_CODE,
      BarcodeFormat.RSS_14,
      BarcodeFormat.RSS_EXPANDED,
      BarcodeFormat.UPC_A,
      BarcodeFormat.UPC_E,
      BarcodeFormat.MSI_PLESSEY
    ];
  }

  static List<AcceptedDocumentFormats> allDocumentFormatTypes() {
    return [
      AcceptedDocumentFormats.AAMVA,
      AcceptedDocumentFormats.BOARDING_PASS,
      AcceptedDocumentFormats.DE_MEDICAL_PLAN,
      AcceptedDocumentFormats.MEDICAL_CERTIFICATE,
      AcceptedDocumentFormats.ID_CARD_PDF_417,
      AcceptedDocumentFormats.SEPA,
      AcceptedDocumentFormats.SWISS_QR,
      AcceptedDocumentFormats.VCARD
    ];
  }

  static List<BarcodeFormat> oneDimensionalBarcodes() {
    return [
      BarcodeFormat.CODABAR,
      BarcodeFormat.CODE_128,
      BarcodeFormat.CODE_39,
      BarcodeFormat.CODE_93,
      BarcodeFormat.EAN_13,
      BarcodeFormat.EAN_8,
      BarcodeFormat.ITF,
      BarcodeFormat.UPC_A,
      BarcodeFormat.RSS_14,
      BarcodeFormat.RSS_EXPANDED,
      BarcodeFormat.UPC_E,
      BarcodeFormat.MSI_PLESSEY
    ];
  }

  ///return list of two dimensional barcode formats
  static List<BarcodeFormat> twoDimensionalBarcodes() {
    return [
      BarcodeFormat.AZTEC,
      BarcodeFormat.DATA_MATRIX,
      BarcodeFormat.PDF_417,
      BarcodeFormat.QR_CODE
    ];
  }
}

enum MSIPlesseyChecksumAlgorithm {
  None,
  Mod10,
  Mod11IBM,
  Mod11NCR,
  Mod1010,
  Mod1110IBM,
  Mod1110NCR
}

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class BarcodeScannerConfiguration {
  /// Background color of the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsColor;

  /// Background color outside the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderTextHintColor;

  bool? cancelButtonHidden;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// cancel scanning and close view on timeout.
  int? autoCancelTimeout;

  /// Aspect ratio of finder view.
  FinderAspectRatio? finderAspectRatio;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Controls whether to play a beep sound after a successful detection.
  /// Default value is TRUE.
  bool? successBeepEnabled;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.

  CameraOrientationMode? orientationLockMode;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat>? barcodeFormats;

  /// Barcode formats which need to be recognized.
  List<AcceptedDocumentFormats>? acceptedDocumentFormats;

  /// Additional parameters for tweaking the detection of barcodes.
  BarcodeAdditionalParameters? additionalParameters;

  /// Param that handle whether we need to save image from camera preview or make a snapshot while snapping barcode
  /// image uri will be returned in [BarcodeScanningResult.snappedImage] property
  BarcodeImageGenerationType barcodeImageGenerationType =
      BarcodeImageGenerationType.NONE;

  /// Type of engine to be used in barcode scanner
  EngineMode engineMode;

  ///Set current zoom by a linear zoom value ranging from 0f to 1.0f. Optical Zoom Level 0f represents the minimum zoom while Optical Zoom Level 1.0f represents the maximum zoom.
  ///Default value is 0f.
  double? cameraZoomFactor;

  /// Locks the minimal supported focus distance and disables a continuous focus
  /// Default value is false. Only Compatible with CameraX in Android
  bool? minFocusDistanceLock;

  BarcodeScannerConfiguration({
    this.topBarBackgroundColor,
    this.topBarButtonsColor,
    this.cameraOverlayColor,
    this.finderLineColor,
    this.finderTextHintColor,
    this.cancelButtonTitle,
    this.enableCameraButtonTitle,
    this.enableCameraExplanationText,
    this.finderTextHint,
    this.autoCancelTimeout,
    this.finderAspectRatio,
    this.finderLineWidth,
    this.successBeepEnabled,
    this.flashEnabled,
    this.orientationLockMode,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
    this.cancelButtonHidden,
    this.barcodeImageGenerationType = BarcodeImageGenerationType.NONE,
    this.additionalParameters,
    this.engineMode = EngineMode.NextGen,
    this.cameraZoomFactor,
    this.minFocusDistanceLock,
  });

  factory BarcodeScannerConfiguration.fromJson(Map<String, dynamic> json) =>
      _$BarcodeScannerConfigurationFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeScannerConfigurationToJson(this);
}

enum BarcodeImageGenerationType { NONE, VIDEO_FRAME, CAPTURED_IMAGE }

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class DetectBarcodeFromImageParams {
  ///file or content uri
  Uri? uri;

  /// list of multiple file or content uri
  List<Uri>? uris;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat> barcodeFormats = <BarcodeFormat>[];

  /// Barcode document formats which need to be recognized.
  List<AcceptedDocumentFormats> acceptedDocumentFormats =
      <AcceptedDocumentFormats>[];

  DetectBarcodeFromImageParams(
    this.uri,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
  );

  /// constructor overloading is acheived by giving it a name
  /// Below consturctor is used for multiple images
  DetectBarcodeFromImageParams.FromMultiple(
    this.uris,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
  );

  factory DetectBarcodeFromImageParams.fromJson(Map<String, dynamic> json) =>
      _$DetectBarcodeFromImageParamsFromJson(json);
  Map<String, dynamic> toJson() => _$DetectBarcodeFromImageParamsToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeFormattedData {
  String? title;
  String? subtitle;

  BarcodeFormattedData({this.title, this.subtitle});

  factory BarcodeFormattedData.fromJson(Map<String, dynamic> json) =>
      _$BarcodeFormattedDataFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeFormattedDataToJson(this);
}

typedef BarcodeDataFormatter = Future<BarcodeFormattedData> Function(
    BarcodeItem barcodeItem);

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class BatchBarcodeScannerConfiguration {
  /// Change data representation for scanned list items

  @JsonKey(ignore: true)
  BarcodeDataFormatter? barcodeFormatter;

  /// The background color of the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsColor;

  /// The color of the titles of all inactive buttons in the top toolbar.

  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsInactiveColor;

  /// The color of barcodes count text
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? barcodesCountTextColor;

  /// The color of barcodes details action
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? detailsActionColor;

  /// The color of barcodes details background
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? detailsBackgroundColor;

  /// The color of barcodes details background
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? detailsPrimaryColor;

  /// The background color outside the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderTextHintColor;

  bool? cancelButtonHidden;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// Text of submit button.
  String? submitButtonTitle;

  /// Text of clear button.
  String? clearButtonTitle;

  /// Text of count text.
  String? barcodesCountText;

  /// Text of fetch state text.
  String? fetchingStateText;

  /// Text when there are  no scanned barcodes yet.
  String? noBarcodesTitle;

  /// Aspect ratio of finder view.
  FinderAspectRatio? finderAspectRatio;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Controls whether to play a beep sound after a successful detection.
  /// Default value is TRUE.
  bool? successBeepEnabled;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  /// Additional parameters for tweaking the detection of barcodes.
  BarcodeAdditionalParameters? additionalParameters;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.

  CameraOrientationMode? orientationLockMode;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat>? barcodeFormats;

  /// Barcode document formats which need to be recognized.
  List<AcceptedDocumentFormats>? acceptedDocumentFormats;

  /// Type of engine to be used in barcode scanner
  EngineMode engineMode;

  ///Set current zoom by a linear zoom value ranging from 0f to 1.0f. Optical Zoom Level 0f represents the minimum zoom while Optical Zoom Level 1.0f represents the maximum zoom.
  ///Default value is 0f.
  // double? cameraZoomFactor;

  /// Locks the minimal supported focus distance and disables a continuous focus
  /// Default value is false. Only Compatible with CameraX in Android
  bool? minFocusDistanceLock;

  BatchBarcodeScannerConfiguration({
    this.topBarBackgroundColor,
    this.topBarButtonsColor,
    this.topBarButtonsInactiveColor,
    this.barcodeFormatter,
    this.barcodesCountTextColor,
    this.detailsActionColor,
    this.detailsBackgroundColor,
    this.detailsPrimaryColor,
    this.cameraOverlayColor,
    this.finderLineColor,
    this.finderTextHintColor,
    this.cancelButtonHidden,
    this.cancelButtonTitle,
    this.enableCameraButtonTitle,
    this.enableCameraExplanationText,
    this.finderTextHint,
    this.submitButtonTitle,
    this.clearButtonTitle,
    this.barcodesCountText,
    this.fetchingStateText,
    this.noBarcodesTitle,
    this.finderAspectRatio,
    this.finderLineWidth,
    this.successBeepEnabled,
    this.flashEnabled,
    this.orientationLockMode,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
    this.additionalParameters,
    this.engineMode = EngineMode.NextGen,
    this.minFocusDistanceLock,
  });

  // this.cameraZoomFactor});

  factory BatchBarcodeScannerConfiguration.fromJson(
          Map<String, dynamic> json) =>
      _$BatchBarcodeScannerConfigurationFromJson(json);

  Map<String, dynamic> toJson() =>
      _$BatchBarcodeScannerConfigurationToJson(this);
}

/// A class that contains additional parameters for barcodes detection and filtering.
@JsonSerializable(includeIfNull: false, explicitToJson: true)
class BarcodeAdditionalParameters {
  /// With this option, the scanner assumes that the barcode can be a GS1 barcode, and modifies the behavior as needed.
  /// You can set it to FALSE, if you don't want to see decoded FNC1 characters ("]C1" and ASCII char 29).
  /// The default is TRUE.
  bool? enableGS1Decoding;

  /// Minimum required text length of the detected barcode.
  /// The default is 0.
  /// NOTE: Currently works for ITF barcodes only!
  int? minimumTextLength;

  /// Maximum text length of the detected barcode. Setting to zero removes the limit.
  /// The default is 0.
  /// NOTE: Currently works for ITF barcodes only!
  int? maximumTextLength;

  /// Minimum required quiet zone on the barcode. Measured in modules (the size of minimal bar on the barcode).
  /// The default is 10.
  /// NOTE: Currently works for ITF barcodes only!
  int? minimum1DBarcodesQuietZone;

  /// The checksum algorithm for MSI Plessey barcodes.
  /// The default value is `Mod10'
  MSIPlesseyChecksumAlgorithm? msiPlesseyChecksumAlgorithm;

  BarcodeAdditionalParameters(
      {this.enableGS1Decoding,
      this.minimumTextLength,
      this.maximumTextLength,
      this.minimum1DBarcodesQuietZone,
      this.msiPlesseyChecksumAlgorithm});

  factory BarcodeAdditionalParameters.fromJson(Map<String, dynamic> json) =>
      _$BarcodeAdditionalParametersFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeAdditionalParametersToJson(this);
}
