import 'package:flutter/cupertino.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:scanbot_sdk/common_data.dart';
import 'package:scanbot_sdk/sdk_utils.dart';

part 'mrz_scanning_data.g.dart';

/// This class describes machine readable zone rtu-ui component scanned result
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class MrzScanningResult extends Result {
  List<MRZField> fields = <MRZField>[];
  int validCheckDigitsCount = 0;
  int checkDigitsCount = 0;

  MrzScanningResult(
      this.fields, this.validCheckDigitsCount, this.checkDigitsCount);

  MrzScanningResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  MrzScanningResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory MrzScanningResult.fromJson(Map<String, dynamic> json) =>
      _$MrzScanningResultFromJson(json);

  Map<String, dynamic> toJson() => _$MrzScanningResultToJson(this);
}

/// This class describes machine readable zone one scanned field data
@JsonSerializable(createFactory: true, includeIfNull: false)
class MRZField {
  final String value;
  final double averageRecognitionConfidence;
  final MRZFieldName name;

  MRZField(this.value, this.averageRecognitionConfidence, this.name);

  factory MRZField.fromJson(Map<String, dynamic> json) =>
      _$MRZFieldFromJson(json);

  Map<String, dynamic> toJson() => _$MRZFieldToJson(this);
}

enum MRZFieldName {
  Unknown,
  DocumentCode,
  IssuingStateOrOrganization,
  DepartmentOfIssuance,
  FirstName,
  LastName,
  Nationality,
  DateOfBirth,
  Gender,
  DateOfExpiry,
  PersonalNumber,
  TravelDocType,
  Optional1,
  Optional2,
  DiscreetIssuingStateOrOrganization,
  KindOfDocument
}

/// This class describes machine readable zone rtu-ui component configuration
@JsonSerializable(includeIfNull: false, explicitToJson: true)
class MrzScannerConfiguration {
  /// Background color of the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsColor;

  /// Background color outside the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderTextHintColor;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// Aspect ratio of finder view.
  FinderAspectRatio? finderAspectRatio;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Controls whether to play a beep sound after a successful detection.
  /// Default value is TRUE.
  bool? successBeepEnabled;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  /// Title of flash button. iOS Only
  String? flashButtonTitle;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.
  CameraOrientationMode? orientationLockMode;

  bool? cancelButtonHidden;

  MrzScannerConfiguration(
      {this.topBarBackgroundColor,
      this.topBarButtonsColor,
      this.cameraOverlayColor,
      this.finderLineColor,
      this.finderTextHintColor,
      this.cancelButtonTitle,
      this.enableCameraButtonTitle,
      this.enableCameraExplanationText,
      this.finderTextHint,
      this.finderAspectRatio,
      this.finderLineWidth,
      this.successBeepEnabled,
      this.flashEnabled,
      this.flashButtonTitle,
      this.orientationLockMode,
      this.cancelButtonHidden});

  factory MrzScannerConfiguration.fromJson(Map<String, dynamic> json) =>
      _$MrzScannerConfigurationFromJson(json);

  Map<String, dynamic> toJson() => _$MrzScannerConfigurationToJson(this);
}
