import 'package:flutter/cupertino.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:scanbot_sdk/common_data.dart';
import 'package:scanbot_sdk/sdk_utils.dart';

import 'common_data.dart';

part 'nfc_reader_data.g.dart';

/// The result of a passport reading operation.
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class NfcPassportReaderResult extends Result {
  /// The first data group parsed from the passports NFC chip.
  NfcDataGroup1? dataGroup1;

  /// The second data group parsed from the passports NFC chip
  NfcDataGroup2? dataGroup2;

  /// The owner's photo image URI.
  Uri? photoUri;

  NfcPassportReaderResult(this.dataGroup1, this.dataGroup2, this.photoUri);

  NfcPassportReaderResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  NfcPassportReaderResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory NfcPassportReaderResult.fromJson(Map<String, dynamic> json) =>
      _$NfcPassportReaderResultFromJson(json);

  Map<String, dynamic> toJson() => _$NfcPassportReaderResultToJson(this);
}

/// Data group 1 or DG1.
/// Contains general information about the passport and its holder.
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class NfcDataGroup1 {
  /// The type of the document.
  String? documentType;

  /// The state or organisation that issued the passport.
  String? issuingStateOrOrganization;

  /// The unique number of the passport.
  String? documentNumber;

  /// The checksum digit of the passports document number.
  int? documentNumberCheckDigit;

  /// Optional data.
  String? optionalData;

  /// Optional data 2.
  String? optionalData2;

  /// Optional data checksum.
  int? optionalDataCheckDigitTd3;

  /// The date of birth of the passports holder.
  @JsonKey(fromJson: dateFromString)
  DateTime? dateOfBirth;

  /// The checksum digit of the date of birth.
  int? dateOfBirthCheckDigit;

  /// The date of expiry of the passport.
  @JsonKey(fromJson: dateFromString)
  DateTime? dateOfExpiry;

  /// The checksum digit of the passports date of expiry.
  int? dateOfExpiryCheckDigit;

  /// The sex of the passport holder.
  String? sex;

  /// The nationality of the passports holder.
  String? nationality;

  /// The composite checksum digit.
  int? compositeCheckDigit;

  /// The name of the passport holder.
  String? nameOfDocumentHolder;

  NfcDataGroup1(
      this.documentType,
      this.issuingStateOrOrganization,
      this.documentNumber,
      this.documentNumberCheckDigit,
      this.optionalData,
      this.optionalData2,
      this.optionalDataCheckDigitTd3,
      this.dateOfBirth,
      this.dateOfBirthCheckDigit,
      this.dateOfExpiry,
      this.dateOfExpiryCheckDigit,
      this.sex,
      this.nationality,
      this.compositeCheckDigit,
      this.nameOfDocumentHolder);

  factory NfcDataGroup1.fromJson(Map<String, dynamic> json) =>
      _$NfcDataGroup1FromJson(json);

  Map<String, dynamic> toJson() => _$NfcDataGroup1ToJson(this);
}

/// Data group 2 or DG2.
/// Contains visual biometric data of the passport holder.
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class NfcDataGroup2 {
  /// The number of images.
  int? numberOfImages;

  /// The version number.
  int? version;

  /// The length of the record.
  int? lengthOfRecord;

  /// The number of facial images.
  int? numberOfFacialImages;

  /// The length of the facial record.
  int? lengthOfFacialRecord;

  /// The number of feature points in the facial record.
  int? numberOfFeaturePoints;

  /// The gender of the holder.
  int? gender;

  /// The eye color of the holder.
  int? eyeColor;

  /// The hair color of the holder.
  int? hairColor;

  /// A mask describing the features.
  int? featureMask;

  /// The expression of the holder.
  int? expression;

  /// The angle of the holders pose.
  int? poseAngle;

  /// The uncertainty of the angle of the holders pose.
  int? poseAngleUncertainty;

  /// The type of the facial image.
  int? faceImageType;

  /// The type of the facial image data.
  int? imageDataType;

  /// The width in pixels of the facial image.
  int? imageWidth;

  /// The height in pixels of the facial image.
  int? imageHeight;

  /// The color space of the facial image.
  int? imageColorSpace;

  /// The source type.
  int? sourceType;

  /// The device type.
  int? deviceType;

  /// The quality of the facial image.
  int? quality;

  NfcDataGroup2(
      this.numberOfImages,
      this.version,
      this.lengthOfRecord,
      this.numberOfFacialImages,
      this.lengthOfFacialRecord,
      this.numberOfFeaturePoints,
      this.gender,
      this.eyeColor,
      this.hairColor,
      this.featureMask,
      this.expression,
      this.poseAngle,
      this.poseAngleUncertainty,
      this.faceImageType,
      this.imageDataType,
      this.imageWidth,
      this.imageHeight,
      this.imageColorSpace,
      this.sourceType,
      this.deviceType,
      this.quality);

  factory NfcDataGroup2.fromJson(Map<String, dynamic> json) =>
      _$NfcDataGroup2FromJson(json);

  Map<String, dynamic> toJson() => _$NfcDataGroup2ToJson(this);
}

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class NfcPassportReaderConfiguration {
  /// Background color of the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsColor;

  /// Background color outside the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? finderTextHintColor;

  /// Foreground color of the flash button when the flash is off.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? topBarButtonsInactiveColor;

  /// The color of the progress bar.
  @JsonKey(fromJson: colorFromString, toJson: toColorString)
  Color? progressBarColor;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Whether the cancel button is hidden or not.
  bool? cancelButtonHidden;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.
  CameraOrientationMode? orientationLockMode;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// String being displayed as a hint to hold device over the NFC chip.
  String? holdOverTheChipTitle;

  /// A hint that notifies about downloading datagroups.
  String? downloadingDataTitle;

  /// A title for the state when there is no NFC chip installed on the device, or it is not available.
  String? noChipErrorTitle;

  /// A title for the error, when NFC reader failed to perform authentication.
  String? authenticationErrorTitle;

  /// A title for the error, when NFC reader failed to download the data.
  String? downloadErrorTitle;

  /// Android only. A title for the retry button.
  String? retryButtonTitle;

  /// Android only. A title for the state when the NFC chip is disabled on the device
  String? chipDisabledErrorTitle;

  /// Android only. A title for the screen of passport scanning.
  String? scanPassportTitle;

  /// Controls whether to play a beep sound after a successful detection.
  /// The default value is TRUE.
  bool? successBeepEnabled;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  /// Defines, if photo image should be stored in internal storage on disk.
  /// If set to FALSE, the result `NfcPassportReaderResult` will not have `photoUri`.
  /// The default value is FALSE.
  bool? shouldSavePhotoImageInStorage;

  NfcPassportReaderConfiguration(
      {this.topBarBackgroundColor,
      this.topBarButtonsColor,
      this.cameraOverlayColor,
      this.finderLineColor,
      this.finderTextHintColor,
      this.orientationLockMode,
      this.cancelButtonHidden,
      this.finderLineWidth,
      this.topBarButtonsInactiveColor,
      this.progressBarColor,
      this.cancelButtonTitle,
      this.enableCameraButtonTitle,
      this.enableCameraExplanationText,
      this.finderTextHint,
      this.holdOverTheChipTitle,
      this.downloadingDataTitle,
      this.noChipErrorTitle,
      this.authenticationErrorTitle,
      this.downloadErrorTitle,
      this.retryButtonTitle,
      this.chipDisabledErrorTitle,
      this.scanPassportTitle,
      this.successBeepEnabled,
      this.flashEnabled,
      this.shouldSavePhotoImageInStorage});

  factory NfcPassportReaderConfiguration.fromJson(Map<String, dynamic> json) =>
      _$NfcPassportReaderConfigurationFromJson(json);

  Map<String, dynamic> toJson() => _$NfcPassportReaderConfigurationToJson(this);
}
