import 'package:json_annotation/json_annotation.dart';

import 'common_data.dart';

part 'ocr_data.g.dart';

///OCR action result
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class OcrResult {
  ///Recognized plain text
  String? plainText;

  ///Optional file URI of generated PDF file (if requested via options).
  String? pdfFileUri;

  /// Array of OCR pages with OCR data. Each OcrPage represents OCR result of given scanned Page document image.
  List<OcrPage> pages = <OcrPage>[];

  OcrResult(this.plainText, this.pdfFileUri, this.pages);

  factory OcrResult.fromJson(Map<String, dynamic> json) =>
      _$OcrResultFromJson(json);

  Map<String, dynamic> toJson() => _$OcrResultToJson(this);
}

/// Represents OCR result of a scanned Page document image.
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class OcrPage {
  /// Recognized words as a list of blocks.
  List<OcrResultBlock> words = <OcrResultBlock>[];

  /// Recognized lines as a list of blocks.
  List<OcrResultBlock> lines = <OcrResultBlock>[];

  /// Recognized paragraphs as a list of blocks.
  List<OcrResultBlock> paragraphs = <OcrResultBlock>[];

  /// Recognized plain text of this OCR page.
  String? text;

  OcrPage(this.text, this.words, this.lines, this.paragraphs);

  factory OcrPage.fromJson(Map<String, dynamic> json) =>
      _$OcrPageFromJson(json);

  Map<String, dynamic> toJson() => _$OcrPageToJson(this);
}

/// An OCR result block represents a recognized paragraph, a line or a word.
/// It contains the recognized plain text, the bounding box and a confidence value.
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class OcrResultBlock {
  /// Recognized plain text of this block.
  String? text;

  /// The bounding box as normalized unit coordinate.
  @JsonKey()
  Rect? boundingBox;

  /// The confidence value describes how confident the OCR engine was when detecting the letters and words.
  // Ranges from 0.0 (not confident) to 100.0 (very confident).
  double? confidenceValue;

  OcrResultBlock(this.text, this.boundingBox, this.confidenceValue);

  factory OcrResultBlock.fromJson(Map<String, dynamic> json) =>
      _$OcrResultBlockFromJson(json);

  Map<String, dynamic> toJson() => _$OcrResultBlockToJson(this);
}

///OCR options
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class OcrOptions {
  ///A list with OCR languages of the text to be recognized (e.g. ["en", "de"]).
  List<String>? languages;

  ///Flag to generate a PDF with OCR layer text (sandwiched PDF).
  bool? shouldGeneratePdf;

  OcrOptions({this.languages, this.shouldGeneratePdf});

  factory OcrOptions.fromJson(Map<String, dynamic> json) =>
      _$OcrOptionsFromJson(json);

  Map<String, dynamic> toJson() => _$OcrOptionsToJson(this);
}

///Sdk OCR configs
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class SdkOcrConfig {
  ///List of languages that installed into sdk (downloaded to languageDataPath, or attached from assets)
  List<String> installedLanguages = <String>[];

  ///Folder where tesseract ocr blobs stores
  String? languageDataPath;

  SdkOcrConfig(this.installedLanguages, this.languageDataPath);

  factory SdkOcrConfig.fromJson(Map<String, dynamic> json) =>
      _$SdkOcrConfigFromJson(json);

  Map<String, dynamic> toJson() => _$SdkOcrConfigToJson(this);
}
