import 'dart:async';
import 'dart:convert';

import 'package:flutter/services.dart';
import 'package:logging/logging.dart';
import 'package:scanbot_sdk/cropping_screen_data.dart';
import 'package:scanbot_sdk/document_scan_data.dart';
import 'package:scanbot_sdk/generic_document_recognizer.dart';
import 'package:scanbot_sdk/license_plate_scan_data.dart';

import 'barcode_scanning_data.dart';
import 'common_data.dart';
import 'ehic_scanning_data.dart';
import 'mrz_scanning_data.dart';
import 'nfc_reader_data.dart';

class ScanbotSdkUi {
  static const MethodChannel _batchBarcodeFormatterCallbackChannel =
      MethodChannel('batch_barcode_data_mapper');
  static const MethodChannel _channel = MethodChannel('scanbot_sdk');

  static Future<String> _invoke(method, Map<String, dynamic> json) async {
    return await _channel.invokeMethod(method, json);
  }

  static bool _wasCanceled(String result) {
    return result == 'CANCELED';
  }

  /// Starts the Ready-to-Use Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBarcodeScanner(
      BarcodeScannerConfiguration config) async {
    try {
      var result = await _invoke('startBarcodeScanner', config.toJson());
      if (_wasCanceled(result)) {
        return BarcodeScanningResult.canceled();
      }
      return BarcodeScanningResult.fromJson(jsonDecode(result));
    } catch (e) {
      return BarcodeScanningResult.error(e.toString());
    }
  }

  static Handler? _handler;

  static void _initBatchFormatterHandler(
      MethodChannel sessionCallbackChannel, BarcodeDataFormatter formatter) {
    _handler = (MethodCall call) async {
      if (call.method == 'getFormattedData') {
        var dataItem =
            await formatter(BarcodeItem.fromJson(jsonDecode(call.arguments)));
        return dataItem.toJson();
      }
    };
  }

  /// Starts the Ready-to-Use Batch Barcode Scanner UI as full-screen modal UI dialog.
  /// The Barcode Scanner is able to recognize 1D- and 2D-barcodes (QR codes, etc).
  static Future<BarcodeScanningResult> startBatchBarcodeScanner(
      BatchBarcodeScannerConfiguration config) async {
    try {
      var json = config.toJson();
      if (config.barcodeFormatter != null) {
        _initBatchFormatterHandler(
            _batchBarcodeFormatterCallbackChannel, config.barcodeFormatter!);
        _batchBarcodeFormatterCallbackChannel.setMethodCallHandler(_handler);
        json['shouldUseBarcodeFormatter'] = true;
      } else {
        _handler = null;
      }

      var barcodeScanningResult =
          await _channel.invokeMethod('startBatchBarcodeScanner', json);
      _handler = null;
      if (_wasCanceled(barcodeScanningResult)) {
        return BarcodeScanningResult.canceled();
      }
      return BarcodeScanningResult.fromJson(jsonDecode(barcodeScanningResult));
    } catch (e) {
      return BarcodeScanningResult.error(e.toString());
    }
  }

  /// Starts the Ready-to-Use Document Scanner UI as full-screen modal UI dialog.
  /// The Document Scanner provides guided, automatic document scanning.
  static Future<DocumentScanningResult> startDocumentScanner(
      DocumentScannerConfiguration config) async {
    try {
      var result = await _invoke('startDocumentScanner', config.toJson());
      if (_wasCanceled(result)) {
        return DocumentScanningResult.canceled();
      }
      return DocumentScanningResult.fromJson(jsonDecode(result));
    } catch (e) {
      return DocumentScanningResult.error(e.toString());
    }
  }

  /// Starts the Ready-to-Use Generic Document Recognizer as a full-screen modal UI dialog.
  /// The Generic Document Recognizer provides guided, automatic document scanning.
  static Future<GenericDocumentRecognizerResult> startGenericDocumentRecognizer(
      GenericDocumentRecognizerConfiguration config) async {
    var result =
        await _invoke('startGenericDocumentRecognizer', config.toJson());
    return GenericDocumentRecognizerResult.fromJson(jsonDecode(result));
  }

  /// Starts the Ready-to-Use MRZ Scanner UI as full-screen modal UI dialog.
  /// The MRZ Scanner is able to recognize and parse data from the Machine Readable Zone
  /// on ID cards, passports, travel documents, etc.
  static Future<MrzScanningResult> startMrzScanner(
      MrzScannerConfiguration config) async {
    try {
      var result = await _invoke('startMrzScanner', config.toJson());
      if (_wasCanceled(result)) {
        return MrzScanningResult.canceled();
      }
      return MrzScanningResult.fromJson(jsonDecode(result));
    } catch (e) {
      return MrzScanningResult.error(e.toString());
    }
  }

  /// Starts the Ready-to-Use EHIC Scanner UI as full-screen modal UI dialog.
  /// The EHIC Scanner is able to recognize and parse data from the backside of
  /// European Health Insurance Cards.
  static Future<HealthInsuranceCardRecognitionResult> startEhicScanner(
      HealthInsuranceScannerConfiguration config) async {
    try {
      var result = await _invoke('startEhicScanner', config.toJson());
      if (_wasCanceled(result)) {
        return HealthInsuranceCardRecognitionResult.canceled();
      }
      return HealthInsuranceCardRecognitionResult.fromJson(jsonDecode(result));
    } catch (e) {
      return HealthInsuranceCardRecognitionResult(
          <HealthInsuranceCardField>[], null);
    }
  }

  /// Starts the Ready-to-Use Cropping UI as full-screen modal UI dialog.
  /// The Cropping UI provides the functionality for manual cropping and rotation of the given [page].
  static Future<CroppingResult> startCroppingScreen(
      Page page, CroppingScreenConfiguration config) async {
    try {
      var arguments = config.toJson();
      arguments['page'] = page.toJson();
      var result = await _invoke('startCroppingScreen', arguments);
      if (_wasCanceled(result)) {
        return CroppingResult.canceled();
      }
      return CroppingResult.fromJson(jsonDecode(result));
    } catch (e) {
      Logger.root.severe(e);
      return CroppingResult.error();
    }
  }

  /// Starts the Ready-to-Use License Plate Scanner UI as full-screen modal UI dialog.
  /// The License Plate Scanner is able to recognize european car license plates.
  static Future<LicensePlateScanResult> startLicensePlateScanner(
      LicensePlateScannerConfiguration config) async {
    try {
      var result = await _invoke('startLicensePlateScanner', config.toJson());
      if (_wasCanceled(result)) {
        return LicensePlateScanResult.canceled();
      }
      return LicensePlateScanResult.fromJson(jsonDecode(result));
    } catch (e) {
      return LicensePlateScanResult.error(e.toString());
    }
  }

  /// Starts the Ready-to-Use NFC passport reader as full-screen modal UI dialog.
  /// The NFC passport reader allows to read passport data from the nfc chip.
  /// The reader will require to read the MRZ data of the passport before the NFC chip reading.
  ///
  /// Note: for iOS devices the iOS13 or newer is required.
  /// For iOS: "Near Field Communication Tag Reading" entitlement is required.
  /// Also for iOS: `info.plist` needs to define the passport nfc application ID,
  /// by adding the "com.apple.developer.nfc.readersession.iso7816.select-identifiers"
  /// entry and setting passport application ID "A0000002471001" as the first element.
  static Future<NfcPassportReaderResult> startNfcPassportReader(
      NfcPassportReaderConfiguration config) async {
    try {
      final json = config.toJson();
      final result = await _invoke('startNfcPassportReader', json);
      if (_wasCanceled(result)) {
        return NfcPassportReaderResult.canceled();
      }
      return NfcPassportReaderResult.fromJson(jsonDecode(result));
    } catch (e) {
      return NfcPassportReaderResult.error(e.toString());
    }
  }
}

typedef Handler = Future Function(MethodCall call);
